/*
 * This file is part of the Kvkbd project.
 * Copyright (C) 2007-2008 Todor Gyumyushev <yodor@developer.bg>
 * Copyright (C) 2008 Guillaume Martres <smarter@ubuntu.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "kbddock.h"

#include <QApplication>
#include <QSettings>
#include <QX11Info>
#include <QPainter>
#include <QWidget>
#include <QDesktopWidget>
#include <QMouseEvent>
#include <QIcon>

#include <X11/Xlib.h>
#include <X11/Xatom.h>

KbdDock::KbdDock(QWidget *parent, Qt::WindowFlags f)
        : DragWidget(parent, f /*Qt::X11BypassWindowManagerHint*/)
{
    window = parent;

    setAttribute(Qt::WA_AlwaysShowToolTips);
    ///setToolTip(tr("Toggle keyboard visibility"));

    setFocusPolicy(Qt::NoFocus);
    setAttribute(Qt::WA_X11DoNotAcceptFocus); //alex
    setWindowFlags(windowFlags() | Qt::WindowDoesNotAcceptFocus); //qt5?

    icn = QPixmap(":/pics/dock.png");
    resize(icn.width(), icn.height());

    setVisible(false);
    moved = false;
}

QWidget* KbdDock::parentWidget() const
{
    return window;
}

void KbdDock::setVisible(bool visible)
{
    if (visible != isVisible()) {
        QWidget::setVisible(visible);
        emit(dockShown(visible));
    }
}

void KbdDock::paintEvent(QPaintEvent *)
{
    QPainter p(this);
    p.drawPixmap(0, 0, icn);
}

KbdDock::~KbdDock()
{
}

void KbdDock::saveState()
{
    QSettings s;
	QRect normalGeom = normalGeometry();

	s.beginGroup("Dock");
	s.setValue("screen",QApplication::desktop()->screenNumber(this));
	s.setValue("x",	normalGeom.x());
	s.setValue("y",	normalGeom.y());
	s.setValue("width",normalGeom.width());
	s.setValue("height",normalGeom.height());
	s.endGroup();
}

void KbdDock::restoreState()
{
    QSettings s;
    s.beginGroup("Dock");

    QRect defaultGeometry(0,0,96,47);
    QRect restoredNormalGeometry;
    int   restoredScreenNumber = s.value("screen",-1).toInt();

    restoredNormalGeometry.setX(s.value("x",          defaultGeometry.x()).toInt());
    restoredNormalGeometry.setY(s.value("y",          defaultGeometry.y()).toInt());
    restoredNormalGeometry.setWidth (s.value("width", defaultGeometry.width()).toInt());
    restoredNormalGeometry.setHeight(s.value("height",defaultGeometry.height()).toInt());

    const QDesktopWidget * const desktop = QApplication::desktop();
    if (restoredScreenNumber >= desktop->numScreens() || restoredScreenNumber<0)
        restoredScreenNumber = desktop->primaryScreen();
    const QRect availableGeometry = desktop->availableGeometry(restoredScreenNumber);

    if (!restoredNormalGeometry.intersects(availableGeometry)) {
        restoredNormalGeometry.moveBottom(qMin(restoredNormalGeometry.bottom(),availableGeometry.bottom()));
        restoredNormalGeometry.moveLeft(  qMax(restoredNormalGeometry.left(),  availableGeometry.left()));
        restoredNormalGeometry.moveRight( qMin(restoredNormalGeometry.right(), availableGeometry.right()));
    }
    restoredNormalGeometry.moveTop(qMax(restoredNormalGeometry.top(), availableGeometry.top()));

    setGeometry(restoredNormalGeometry);
    icn=icn.scaled(QSize(restoredNormalGeometry.width(),restoredNormalGeometry.height()));
/*
    unsigned int v = s.value("dockOpacity",50).toInt()/100.*0XFFFFFFFF;
    XChangeProperty(QX11Info::display(), winId(), 
                XInternAtom (QX11Info::display(), "_NET_WM_WINDOW_OPACITY", false), 
                XA_CARDINAL, 32, PropModeReplace, (unsigned char *)(&v), 1L);
or in Qt:*/
    setWindowOpacity(s.value("dockOpacity",50).toDouble()/100.0);

    s.endGroup();
}

void KbdDock::mouseMoveEvent(QMouseEvent *e)
{
    if (!moved && isDragged()) {
        moved = true;
    }
    DragWidget::mouseMoveEvent(e);
}
void KbdDock::mouseReleaseEvent(QMouseEvent *e)
{
    QWidget *pw = window;

    if (moved) {
        DragWidget::mouseReleaseEvent(e);
        moved = false;
        return;
    }

    if (pw->isVisible()) {
        pw->hide();
    } else {
        pw->show();
    }
    DragWidget::mouseReleaseEvent(e);
}
