#include <string.h>

#include <QDir>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/XKBlib.h>

#include "libflycore.h"


Window FindToplevelWindowByName(Display *dpy,Window win,const char *winName)
{
  Window w=None, rootW, parentW, *children=nullptr;
  unsigned int nChildren=0;

  if (!XQueryTree(dpy, win, &rootW, &parentW, &children, &nChildren)) return w;

  //WM_NAME first
  for (unsigned int i = 0; i < nChildren && w==None; i++)
  {
    XTextProperty xtp;
  if (XGetWMName(dpy, children[i], &xtp))
  {
    char** cl = nullptr;
    int n;
    if (XmbTextPropertyToTextList(dpy, &xtp, &cl, &n)==Success && cl)
    {
      if (strcasecmp(winName, cl[0])==0) w = children[i];
      XFreeStringList(cl);
    }
    if (xtp.value) XFree(xtp.value);
  }
  }

  //then try class name 
  for (unsigned int i = 0; i < nChildren && w==None; i++)
  {
      XClassHint wm_class;        // STRING
      wm_class.res_class = wm_class.res_name = nullptr;

      if (XGetClassHint(dpy, children[i], &wm_class))
      {
          if ((wm_class.res_class && strstr(wm_class.res_name, winName)==wm_class.res_name) ||
              (wm_class.res_name  && strstr(wm_class.res_class, winName)==wm_class.res_class)
             ) w=children[i];
          if (wm_class.res_class) XFree(wm_class.res_class);
          if (wm_class.res_name)  XFree(wm_class.res_name);
      }
  }

  //let check WM_ICON_NAME too
  for (unsigned int i = 0; i < nChildren && w==None; i++)
  {
      XTextProperty xtp;
      if (XGetWMIconName(dpy, children[i], &xtp))
      {
          char** cl = nullptr;
          int n;
          if (XmbTextPropertyToTextList(dpy, &xtp, &cl, &n)==Success && cl)
          {
              if (strcasecmp(winName, cl[0])==0) w = children[i];
              XFreeStringList(cl);
          }
          if (xtp.value) XFree(xtp.value);
      }
  }

  if (nChildren) XFree(children);
  return w;
}

const std::string &GetFlyDir(FlyDir d)
{
    static std::string appSharedDir, themeDir, baseDir, decorDir, empty;

    switch (d) 
    {
       case APP_SHARED_DIR: 
           if (appSharedDir.empty()) appSharedDir = std::string("/usr/share/fly"); 
           return appSharedDir;

        case USER_THEME_DIR:
           if (themeDir.empty()) {
                char *kioskRcEnv = getenv("FLY_RC_FILE"); //if set then theme must be read only
                if (kioskRcEnv && *kioskRcEnv && strstr(kioskRcEnv,"/usr/share/fly-wm"))
                     themeDir = std::string("/usr/share/fly-wm/theme");
                else themeDir = (QDir::homePath()+"/.fly/theme").toStdString();
           }
           return themeDir;

       case USER_BASE_DIR:
           if (baseDir.empty()) baseDir = (QDir::homePath() + "/.fly").toStdString(); 
           return baseDir;

       case WM_DECOR_DIR:
           if (decorDir.empty()) decorDir = std::string("/usr/share/fly-wm/decorations"); 
           return decorDir;

       default: break;
    }
    return empty;
}

const char *flySessionName()
{
  char *tmp = getenv("DESKTOP_SESSION");
  return tmp;
}

std::string flySessionConfigSuffix(WhatFlySession stype)
{
  std::string result;

  switch (stype)
  {
      case FLY_FAILSAFE: //fake case
      case FLY_WESTON:   //fake case
      case FLY_DESKTOP: break;
      case FLY_MINI:    result = ".fly-mini";   break;
      case FLY_TABLET:  result = ".fly-tablet"; break;
      case FLY_MOBILE:  result = ".fly-mobile"; break;
      case FLY_KIOSK:   result = ".fly-kiosk";  break;
      case FLY_TABLET_KIOSK: result = ".fly-tablet-kiosk";  break;
      default:
        const char *tmp = flySessionName();
        if (tmp && strcmp(tmp, "fly")!=0)
        {
            result = ".";
            result += tmp;
        }
        break;
  }
  return result;
}

WhatFlySession whatFlySession()
{
 WhatFlySession ret = FLY_UNKNOWN;

 const char *tmp = flySessionName();

 if (tmp)
 {
   if (!strncasecmp(tmp,"fly",3))
   {
          if (!strcasecmp(tmp, "fly"))        ret=FLY_DESKTOP;
     else if (!strcasecmp(tmp, "fly-tablet")) ret=FLY_TABLET;
     else if (!strcasecmp(tmp, "fly-mini"))   ret=FLY_MINI;
     else if (!strcasecmp(tmp, "fly-mobile")) ret=FLY_MOBILE;
     else if (!strcasecmp(tmp, "fly-weston")) ret=FLY_WESTON;
     else if (!strcasecmp(tmp, "fly-kiosk"))  ret=FLY_KIOSK;
     else if (!strcasecmp(tmp, "fly-tablet-kiosk"))  ret=FLY_TABLET_KIOSK;
   }
   else if (!strcasecmp(tmp, "failsafe"))   ret=FLY_FAILSAFE;
 }
 return ret;
}

int GetStandardIconSize(int itemSize)
{
    int  iconSize;
         if (itemSize < 22) iconSize = 16;
    else if (itemSize < 32) iconSize = 22;
    else if (itemSize < 48) iconSize = 32;
    else if (itemSize < 64) iconSize = 48;
    else if (itemSize < 96) iconSize = 64;
    else if (itemSize < 128)iconSize = 96;
    else if (itemSize < 192)iconSize = 128;
    else if (itemSize < 256)iconSize = 192;
    else                    iconSize = 256;
    return iconSize;
}

