/*
 * This file is part of the Kvkbd project.
 * Copyright (C) 2007-2008 Todor Gyumyushev <yodor@developer.bg>
 * Copyright (C) 2008 Guillaume Martres <smarter@ubuntu.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include <QDebug>

#include <QByteArray>
#include <QX11Info>
#include <QTimer>
#include <QMouseEvent>
#include <QProcess>
#include <QFileInfo>

#include "main.h"
#include "mainwidget.h"

#include <QCommandLineParser>
#include <flyintegration.h>
#ifdef WITH_FLY_CORE
#include <flyde.h>
#include <flymisc.h>
#include <flyfileutils.h>
#include <fly/SimpleIni.h>
#else
#include "SimpleIni.h"
#include "libflycore.h"
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/XKBlib.h>

#define explicit dont_use_cxx_explicit
#include <xcb/xkb.h>
#undef explicit

bool keyOnButtonRelease=false;

QCommandLineParser args;

static QString GetStyle(const QString styleIn)
{
    QString style = styleIn;

    if (style!="restore") 
    {
        if (style.isEmpty()) style = "default_flat";

        QString styleFilePath = style;
        if (!styleFilePath.startsWith('/'))
            styleFilePath = QString::fromStdString(GetFlyDir(APP_SHARED_DIR))+"/data/fly-vkbd/styles/"+style+".css";

        if (QFileInfo(styleFilePath).exists())
        {
            QFile styleFile(styleFilePath);
            if (styleFile.open(QIODevice::ReadOnly | QIODevice::Text))
            {
              style.clear();
              while (!styleFile.atEnd()) {
                  QByteArray line = styleFile.readLine();
                  if (!line.startsWith('#') && !line.startsWith("//") && !line.startsWith("/*"))
                      style += QString(" ") + QString::fromLocal8Bit(line);
              }
            }
        }
    }
    return style;
}


int main(int argc, char **argv)
{
/*no effect
    QCoreApplication::setAttribute(Qt::AA_Use96Dpi,true);
    QCoreApplication::setAttribute(Qt::AA_UseHighDpiPixmaps,true);
*/
	/*KAboutData about("kvkbd", 0, ki18n("Kvkbd"), version, ki18n(description),
	                 KAboutData::License_GPL, ki18n("(C) 2007-2008 The Kvkbd Developers"));
	about.addAuthor(ki18n("Todor Gyumyushev"), ki18n("Original Author"), "yodor@developer.bg");
	about.addAuthor(ki18n("Guillaume Martres"), ki18n("KDE4 port"), "smarter@ubuntu.com");
	about.setProgramIconName("preferences-desktop-keyboard");*/
	KvkbdApp app(argc, argv);

	flyInit(
#ifdef SOURCE_VERSION
	SOURCE_VERSION,
#else
	"2.0.0",
#endif
	 QT_TRANSLATE_NOOP("Fly", "A virtual keyboard"));

    QStringList paths; paths << "/usr/share/fly/data/fly-vkbd/icons";
    QIcon::setFallbackSearchPaths(paths);

	app.setWindowIcon(QIcon::fromTheme("input-keyboard"));
/*
	FlyCmdLineArgs::init(argc, argv);
	FlyCmdLineOptions options;

	options.add("loginhelper", QObject::tr("Stand alone use with fly-dm, kdm, etc.\nAdd 'HOME=/root fly-vkbd --loginhelper --delay 2 &' to Xsetup"));
	options.add("strut", QObject::tr("Take uncoverable place at the edge of the screen, like docked panel"));
	options.add("strut-bottom", QObject::tr("Take uncoverable place at the bottom edge of the screen, like docked panel"));
	options.add("minimizetotray", QObject::tr("Start minimized to tray"));
	options.add("delay <secs>", QObject::tr("Raise window delay in seconds"), "0");
	options.add("s <desc>", QObject::tr("Style, use \"restore\" to restore previous style"),"");
	
	FlyCmdLineArgs::addCmdLineOptions(options);
	FlyCmdLineArgs *args = FlyCmdLineArgs::parsedArgs();
*/
    args.addHelpOption(); //dont forget to check if conflict with custom -h option
    //args.addOption(QCommandLineOption("loginhelper", QObject::tr("Stand alone use with fly-dm, kdm, etc.\nAdd 'HOME=/root fly-vkbd --loginhelper --delay 2 &' to Xsetup")));
    args.addOption(QCommandLineOption("strut",         QObject::tr("Take uncoverable place at the edge of the screen, like docked panel")));
    args.addOption(QCommandLineOption("strut-bottom",  QObject::tr("Take uncoverable place at the bottom edge of the screen, like docked panel")));
    args.addOption(QCommandLineOption("minimizetotray",QObject::tr("Start minimized to tray")));
    args.addOption(QCommandLineOption("delay",         QObject::tr("Raise window delay in seconds"),"delay", "0"));
    args.addOption(QCommandLineOption("s",             QObject::tr("Style, css style string, file path or keyword \"restore\" to restore previous style"),"style"));
    args.addOption(QCommandLineOption("color",         QObject::tr("Color, basic color"),"color","default"));
    args.addOption(QCommandLineOption("update",        QObject::tr("Update view")));
    args.addOption(QCommandLineOption("toggle",        QObject::tr("Toggle view")));
    args.addOption(QCommandLineOption("geom",          QObject::tr("Geometry as in X11 applications"), "geom", "600x200-0-0"));
    args.addOption(QCommandLineOption("keyonrelease",  QObject::tr("Send key when mouse button pressed or released, default - when pressed")));
    args.process(app);
    //for (int i=0; i< argc; i++) fprintf(stderr,"vkbd argument[%d]=%s gm=%s\n",i,argv[i]);

    // FIXME: kvkbd is hidden by KDM4
    // In loginhelper mode, settings can't be changed and there's no tray icon
    
    /*bool alone = false;
    if (args.isSet("loginhelper")) {
    alone = true;
    }
    else*/ if (app.isRunning())
    {
        if (args.isSet("update")) { 
            app.sendMessage("update");
            return -1;
        }

        QString styleArg = args.value("s");
        if (!styleArg.isEmpty())
        {
            QString style = GetStyle(styleArg); 
            app.sendMessage("style:"+style);
        }
     
        if (args.isSet("geom"))
        {
            QString geomArgString = args.value("geom");
            if (!geomArgString.isEmpty()) app.sendMessage("geom:"+geomArgString);
        }
             
        if (args.isSet("toggle"))
        {
            app.sendMessage("toggle");
            return -1;
        }

        if (!args.isSet("minimizetotray"))
             app.sendMessage("raise");
        else app.sendMessage("minimizetotray");
        
        return -1;
    }
    else {
        if (args.isSet("update")) return -1;
    }

    if (args.isSet("keyonrelease")) keyOnButtonRelease=true;

	app.initialize();

    app.setMainWindow(/*alone*/);

	
	// TODO: find a better way to detect if we're launched from a display manager
	/*TODO    KUser user;
	    if (alone && !user.isSuperUser()) {
	        qCritical() << "LoginHelper mode can only be launched from a display manager!";
	        return -1;
	    }*/

	return app.exec();
}

KvkbdApp::KvkbdApp(int& argc, char** argv)
	: QtSingleApplication(QString::fromLocal8Bit(getenv("DISPLAY"))+"fly-vkbd",argc, argv)
{
    mainWindow = NULL;

	//QAbstractEventDispatcher::instance()->setEventFilter(allEventsFilter);
	int opcode_rtrn, error_rtrn;
	int major = XkbMajorVersion;
	int minor = XkbMinorVersion;

	if (!XkbLibraryVersion(&major, &minor))
		qDebug("Incompatible  Xlib XKB extension v.%d.%d!=%d.%d - try to rebuild this program\n",
		       major, minor, XkbMajorVersion, XkbMinorVersion);
	else if (!XkbQueryExtension(QX11Info::display(), &opcode_rtrn, &xkbEventId, &error_rtrn, &major, &minor))
		qDebug("Incompatible or disabled X server XKB extension - check X server\n");
	else
	{
		//it requires additional libxkbui library:
		//::XkbInitAtoms(NULL);// Do it, or face horrible memory corrupting bugs
		//qDebug() << "xkbEventId = " << xkbEventId;
		bool useXkbExt = XkbSelectEventDetails(QX11Info::display(), XkbUseCoreKbd, XkbStateNotify,
		                                       XkbAllStateComponentsMask, XkbGroupStateMask);

		if (!useXkbExt)
			qDebug("Not using xkb ext");
	}

	setQuitOnLastWindowClosed(false); //alex: not to quit when close settings dialog by fly-vkbd hidden in tray
	
	connect(this, SIGNAL(aboutToQuit()), this, SLOT(closeAllWindows()));
	connect(this, SIGNAL(messageReceived(QString)), this, SLOT(processAppMessage(QString)));


	XcbEventFilter *ef = new XcbEventFilter(this);
	installNativeEventFilter(ef);
}

KvkbdApp::~KvkbdApp()
{
}

void KvkbdApp::saveState(QSessionManager &sm)
{
    sm.setRestartHint(QSessionManager::RestartNever);
}

typedef union {
        /* All XKB events share these fields. */
        struct {
            uint8_t response_type;
            uint8_t xkbType;
            uint16_t sequence;
            xcb_timestamp_t time;
            uint8_t deviceID;
        } any;
        xcb_xkb_map_notify_event_t map_notify;
        xcb_xkb_state_notify_event_t state_notify;
} _xkb_event;

// update keyboard mapping when it's changed by xmodmap/setxkbmap
bool KvkbdApp::xcbEventFilter(const QByteArray &eventType, void *message, long *d)
{
	if (!mainWindow) return false;

	Window mW = mainWindow->winId();
	if (!mW) return false;

	xcb_generic_event_t* event = static_cast<xcb_generic_event_t *>(message);
	unsigned int responce_type = event->response_type & ~0x80;

	
	if (responce_type == XCB_VISIBILITY_NOTIFY)
	{
      if ((((xcb_visibility_notify_event_t *)event)->window)==mW)
	  {
		  if (((xcb_visibility_notify_event_t *)event)->state != XCB_VISIBILITY_UNOBSCURED)
			mainWindow->raise();
	  }
	}

	/*XkbEvent* ev = (XkbEvent*)event;
	if (ev->type == xkbEventId && ev->any.xkb_type == XkbStateNotify)
	{
		XkbStateNotifyEvent* snEv = (XkbStateNotifyEvent*)ev;
		mainWindow->mappingNotify(snEv->group);
	}*/
	if (responce_type == xkbEventId)
	{
		_xkb_event * xkbEvent = reinterpret_cast<_xkb_event *>(event);

		if (xkbEvent && xkbEvent->any.xkbType==XCB_XKB_STATE_NOTIFY)
		  mainWindow->mappingNotify(xkbEvent->state_notify.group);

	}

	return false;
}

// FIXME: What's this for?
//  int newInstance(){
//   MainWidget *main = (MainWidget *)mainWidget();
//   if (!main)
//   {
//    main = new MainWidget(const_cast<KAboutData *>(aboutData()),false, 0, "kvkbd");
//    setMainWidget(main);
//   }
//   return 0;
//  };

/*bool KvkbdApp::allEventsFilter(void * message)
{
	XEvent *event = static_cast<XEvent*>(message);
	qDebug() << Q_FUNC_INFO << event->type;
	MainWidget *main = (MainWidget *)activeWindow();
	if (main) { //FIXME: is this check really necessary?
		if (event->type == MappingNotify) {
			qDebug() << "mapping event";
			XMappingEvent *e = (XMappingEvent *)event;
			if (e->request == MappingKeyboard) {
				main->mappingNotify(e);
			}
		}
	}
	return false;
}
*/

void KvkbdApp::setMainWindow()
{
    bool firstTime = mainWindow?false:true;

    if (mainWindow) { delete mainWindow; mainWindow=NULL; }

    //get base color ---------------------------
    QColor color;
	QString colorString = args.value("color");
	if (colorString.isEmpty() || colorString=="default")
	{
    QString themePath = QString::fromStdString(GetFlyDir(USER_THEME_DIR))+"/"+
                        QString::fromLocal8Bit(FLY_CURRENT_THEME_NAME) + 
                        QString::fromStdString(flySessionConfigSuffix());

    CSimpleIniA ini(true,false,false);
    SI_Error rc = ini.LoadFile(themePath.toLocal8Bit().data());
    if (rc>=0)
    {
        const char *val = ini.GetValue("Variables", "DecorTheme");
        if (!val || !*val) //no decor => take color from theme
        {
            const char *val = ini.GetValue("Variables", "TaskbarColor");
            if (val && *val) 
            {
                if (!strcmp(val,"qt"))
                     color = QApplication::style()->standardPalette().color(QPalette::Background);
                else
                {
                    QString themePath = QString::fromStdString(GetFlyDir(USER_BASE_DIR))+"/paletterc";
                    CSimpleIniA ini(true,false,false);
                    SI_Error rc = ini.LoadFile(themePath.toLocal8Bit().data());
                    if (rc>=0)
                    {
                        val = ini.GetValue("Variables", "PrimaryColor");
                        if (val && *val)
                        {
                            if (!strcasecmp(val,"qt"))
                                 color = QApplication::style()->standardPalette().color(QPalette::Background);
                            else
                            {
                                if (QColor::isValidColor(QString::fromLatin1(val))) color = QColor(val);
                                else
                                {
                                    fprintf(stderr,"can't get color from fly-wm rc files\n");
                                    color = QApplication::style()->standardPalette().color(QPalette::Background);
                                }
                            }
                        }
                    }
                }
            }
        }
        else //take color from taskbar bg image
        {
            themePath =  QString::fromStdString(GetFlyDir(WM_DECOR_DIR)) + "/" + QString::fromLocal8Bit(val) + "/taskbar/taskbarbg.xpm";
            QImage taskbarImage(themePath);
            if (!taskbarImage.isNull())
            {
                QRgb pix = taskbarImage.pixel(0,0);
                color = QColor(pix);
            }
        }
    }
	}
	else
	{
		color = QColor(colorString);
	}
    //-------------------------------------------

    MainWidget::StrutType strut = MainWidget::StrutNone;
    	 if (args.isSet("strut"))			strut = MainWidget::StrutCurrent;
    else if (args.isSet("strut-bottom"))	strut = MainWidget::StrutBottom;

    QString styleArg = args.value("s");
    QString style = GetStyle(styleArg);

    mainWindow = new MainWidget(strut, color, style/*, alone*/);
    
    mainWindow->setInLocker(
    	FindToplevelWindowByName(QX11Info::display(),QX11Info::appRootWindow(),FLYWM_LOCKER_WINDOW_NAME) ||
    	FindToplevelWindowByName(QX11Info::display(),QX11Info::appRootWindow(),"fly-dm_greet") 
    	);

    if (args.isSet("geom")) mainWindow->setArgGeometry(args.value("geom"));

    if (!args.isSet("minimizetotray")) {
    	mainWindow->show();
    }
    else if (firstTime) {
        KbdTray *trayIcon = mainWindow->kbdTray();
        if (trayIcon) trayIcon->showMessage(tr("Virtual keyboard"), tr("Started"), QSystemTrayIcon::Information, 2000);
    }

    if (args.isSet("delay"))
    {
    	int raiseAfter = args.value("delay").toInt();
    	QTimer::singleShot(raiseAfter*1000, mainWindow, SLOT(raise()));
    }
}

void KvkbdApp::processAppMessage(QString message)
{
    if (message=="quit")   { quit(); return; }

    if (message=="update") { setMainWindow(); return; }

    bool wasShown=mainWindow->isVisible(), inLocker=false;
    if (mainWindow)
    {
        if (message.startsWith("style:")) mainWindow->setCssStyle(message.mid(6));


        inLocker = (FindToplevelWindowByName(QX11Info::display(),QX11Info::appRootWindow(),FLYWM_LOCKER_WINDOW_NAME) ||
                        FindToplevelWindowByName(QX11Info::display(),QX11Info::appRootWindow(),"fly-dm_greet"));
        mainWindow->setInLocker(inLocker);

        if (message == "raise")			  {
            mainWindow->setVisible(true);
            mainWindow->raise();
        }
        if (message == "minimizetotray")  mainWindow->hide();

        if (message == "toggle") mainWindow->setVisible(!mainWindow->isVisible());

        if (message.startsWith("geom:")) mainWindow->setArgGeometry(message.mid(5));
    }

    if (!inLocker && !wasShown && !mainWindow->isVisible()) {
        KbdTray *trayIcon = mainWindow->kbdTray();
        if (trayIcon) trayIcon->showMessage(tr("Virtual keyboard"), tr("Started"), QSystemTrayIcon::Information, 2000);
    }
}

