/*
 * This file is part of the Kvkbd project.
 * Copyright (C) 2008 Guillaume Martres <smarter@ubuntu.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef MAIN_H
#define MAIN_H

#include <qtsingleapplication.h>

#include <QAbstractNativeEventFilter>

#include <QSessionManager>

extern bool keyOnButtonRelease;

class MainWidget;
class QCommandLineParser;

class KvkbdApp : public QtSingleApplication
{
Q_OBJECT
public:
	explicit KvkbdApp(int& argc, char** argv);
	virtual ~KvkbdApp();

    virtual bool xcbEventFilter(const QByteArray &eventType, void *message, long *d);

	void setMainWindow();

public slots:
    virtual void saveState(QSessionManager &sm);

private slots:
	void processAppMessage(QString message);

private:
	int xkbEventId;
	MainWidget* mainWindow;
};

class XcbEventFilter : public QAbstractNativeEventFilter
{
    KvkbdApp *app;

public:
    XcbEventFilter(KvkbdApp *a) : app(a) {}

    virtual bool nativeEventFilter(const QByteArray &eventType, void *message, long *d) Q_DECL_OVERRIDE
    {
        if (eventType == "xcb_generic_event_t")
        {
          return app->xcbEventFilter(eventType,message,d);
        }
        return false;
    }
};

#endif //MAIN_H
