/*
 * This file is part of the Kvkbd project.
 * Copyright (C) 2007-2008 Todor Gyumyushev <yodor@developer.bg>
 * Copyright (C) 2008 Guillaume Martres <smarter@ubuntu.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */


#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include <X11/XKBlib.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <X11/extensions/XKBrules.h>
#include "fixx11h.h"

#include <QDebug>

#include <QApplication>
#include <QProcess>
#include <QResizeEvent>
#include <QX11Info>
#include <QDesktopWidget>
#include <QHash>
#include <QTimer>
//#include <QPainter>
#include <QMenu>
#include <QAction>
#include <QFontDialog>
#include <QColorDialog>
#include <QSettings>
#include <QTimer>
#include <QMessageBox>
#include <QScreen>
#include <QWindow>
#include <QDir>
#ifdef WITH_SONNET
//#include <QTableWidget>
#include <QListWidget>
#include <QHeaderView>
#include <QScroller>
#include <QTouchDevice>
#include <QToolButton>
#include <QStyledItemDelegate>
#define FLY_VKBD_LIST_WIDGET_NAME "FlyVkbdListWidget"
#endif

//#ifndef NO_KWINDOWSYSTEM
#include <KWindowSystem>
//#endif

#ifdef WITH_SOLID
#include <Solid/Device>
#include <Solid/DeviceNotifier>
using namespace Solid;
#endif
#ifdef QAccessibilityClient_FOUND
#include <QAccessible>
#include <qaccessibilityclient/accessibleobject.h>
using namespace  QAccessibleClient;
#endif

#include <flyintegration.h>
#include <flyabout.h>
#include <flyhelpmenu.h>
#ifdef WITH_FLY_CORE
#include <flyicontheme.h>
#include <flyfileutils.h>
#include <flyde.h>
#else
#include "libflycore.h"
#endif
#include <flyicon.h>

#include "main.h"
#include "mainwidget.h"

#include "kbddock.h"
#include "vbutton.h"
#include "controlbutton.h"
#include "numpadvbutton.h"
#include <keysym2ucs.h>

#define ROWLENGTH1 13
#define ROWLENGTH2 10
#define ROWLENGTH3 9
#define ROWLENGTH4 7

static const int hKeySpace = 3;
static const int vKeySpace = 3;

static const int stx = 5;//15;// alex - if change - to much another changes
static const int sty = 5;

VButton *globalShift=NULL;
VButton *rightShift=NULL;
#ifdef WITH_SONNET
bool    useSonnet=true;
QString *currentString=NULL;
#define MAX_LANG 4
#include "rus2lat.h"
#endif
bool autoRepeated=true;
bool processRightButton=false;
bool fixedRightShift=true;
int  vkbdAutoRepeatDelay=660, vkbdAutoRepeatRate=40;

QIcon resizeIcon;//shrinkIcon, expandIcon;

static QString fontWeightString(int weight)
{
	QString ws;
	switch (weight)
	{
	case QFont::Light:		ws="light";    break;
	case QFont::DemiBold:	ws="demibold"; break;
	case QFont::Bold:		ws="bold";     break;
	case QFont::Black:		ws="black";    break;
	default:
	case QFont::Normal:		ws="normal";   break;
	}
	return ws;
}

#ifdef WITH_SONNET
MyListWidget::MyListWidget(QSize shrinkSize, QSize expandedSize,QPoint pos,QWidget *p) : QListWidget(p)
{
    shrinked=true;
    sSize=shrinkSize;
    eSize=expandedSize;

    setAccessibleName(FLY_VKBD_LIST_WIDGET_NAME);
    setFocusPolicy(Qt::NoFocus);

    setViewportMargins(0,0,shrinkSize.height(),0);

    //if (shrinkIcon.isNull()) shrinkIcon = QIcon::fromTheme("go-up");
    //if (expandIcon.isNull()) expandIcon = QIcon::fromTheme("go-down");
    if (resizeIcon.isNull()) resizeIcon = QIcon::fromTheme(/*"view-fit-height"*/"view-fullscreen");

    shrinkExpandBtn = new QToolButton(this);
    shrinkExpandBtn->setIcon(/*expandIcon*/resizeIcon);
    shrinkExpandBtn->resize(QSize(shrinkSize.height()-1,shrinkSize.height()-1));
    shrinkExpandBtn->move(QPoint(shrinkSize.width() - shrinkExpandBtn->width() - 1, 0));
    shrinkExpandBtn->setFocusPolicy(Qt::NoFocus);

    connect(shrinkExpandBtn,SIGNAL(clicked(bool)),this,SLOT(shrinkExpandSlot()));

    int standardIconSize = GetStandardIconSize(shrinkExpandBtn->width()-3);
    shrinkExpandBtn->setIconSize(QSize(standardIconSize,standardIconSize));

    //resize(shrinkSize);
    //move(pos);
    setGeometry(pos.x(),pos.y(),shrinkSize.width(),shrinkSize.height());
}
MyListWidget::~MyListWidget()
{
}

void MyListWidget::shrinkExpandSlot()
{
    resize(shrinked?eSize:sSize);

    shrinked = !shrinked;

    //shrinkExpandBtn->setIcon(shrinked?expandIcon:shrinkIcon);

    raise();
}
#endif

MainWidget::MainWidget(StrutType s, const QColor &color, const QString& styleSheet, //bool loginhelper,
                       QWidget *parent, Qt::WindowFlags f)
    : ResizableDragWidget(parent, s ? Qt::Tool|Qt::WindowStaysOnTopHint|Qt::CustomizeWindowHint : f), strut(s)
{
    dpiMul = 1.0;
    sdxs = sdxb = 0.0;

    bgColor = color;
    if (bgColor.isValid())
    {
        if (bgColor.name()=="#000000")
             bgColor  = "#222222";
        else bgColor = bgColor.darker(120);
    }
    else bgColor = QColor("#7Db8C2");

#ifdef WITH_SONNET
    speller         = NULL;
    suggestMenu     = NULL;
    suggestTable    = NULL;
    fromClipboard   = false;
    currentString   = NULL;
    //guesser       = NULL;

    QSettings cfg;
    useSonnet       = cfg.value("useSonnet", false).toBool();
    if (useSonnet)
    {
        speller = new Sonnet::Speller();
        //fprintf(stderr,"speller default lang=%s\n",speller->defaultLanguage().toLocal8Bit().data());

        //guesser = new Sonnet::GuessLanguage();
        //guesser->setLimits(MAX_LANG,0.1);

        currentString = new QString();

        clipboard = QApplication::clipboard();
        connect(clipboard,SIGNAL(changed(QClipboard::Mode)),this,SLOT(checkClipboard(QClipboard::Mode)));

        newIcon = QIcon::fromTheme("list-add");
    }
#endif

    char *e, *dpi_str = XGetDefault(QX11Info::display(),"Xft","dpi");
    if (dpi_str)
    {
          double d = strtod(dpi_str,&e);
          if (e != dpi_str) dpiMul = 96.0/d;
    }
    dpiMul *= devicePixelRatioF();

    visibilityForced=false;

    chooseFontAction=/*chooseBackgroundColorAction=*/helpAction=NULL;


    m_XkbDescPtr = NULL;

    int opcode_rtrn, error_rtrn;
    int major = XkbMajorVersion;
    int minor = XkbMinorVersion;
    int xkbEventId;
    if (!XkbQueryExtension(QX11Info::display(), &opcode_rtrn, &xkbEventId, &error_rtrn, &major, &minor))
    {
        fprintf(stderr,"fly-vkbd: Incompatible or disabled X server XKB extension - check X server\n");
    }
    else
    {
        m_XkbDescPtr = getXkbDesc();
    }
    if (!m_XkbDescPtr)
    {
        if (QMessageBox::question(this,tr("Attention"),tr("XKEYBOARD (XKB) extension failed! Continue?"))!=QMessageBox::Yes) 
          exit(-1);
    }

    if (!styleSheet.isEmpty()) cssStyle = styleSheet;

    if (!cssStyle.contains("gradient")) QIcon::setThemeName("fly-astra-flat");


    setAttribute(Qt::WA_AlwaysShowToolTips);
    setFocusPolicy(Qt::NoFocus);
    setAttribute(Qt::WA_X11DoNotAcceptFocus); //alex
    setWindowFlags(windowFlags() | Qt::WindowDoesNotAcceptFocus); //qt5?
    setWindowTitle("fly-vkbd");

    //alone = loginhelper;

    funcRow=false;
    autoShow=true;

    quit = buttonMenu = extent = layoutSwitcher=NULL;
    menu = NULL;
    tray = NULL;
    dock = NULL;
    resizable = true;
    // aboutData = about;
    if (flyInTabletMode())
         strutBottomHeight=0.4;  //default for table mode
    else strutBottomHeight=0.33; //default is 1/3 of screen

    display = QX11Info::display();
    desktop = QApplication::desktop();
    screenNum  = desktop->screenNumber(this);

    //screenGeometry = desktop->screenGeometry();
    QDesktopWidget *dw = QApplication::desktop();
    screenGeometry = dw->screenGeometry(dw->primaryScreen());

/*
    for (int i=0; i<desktop->numScreens();i++)
    {
        const QRect availableGeometry = desktop->availableGeometry(i);
        fprintf(stderr,"screen[%d]= %d %d %d %d\n",i,
                availableGeometry.x(),availableGeometry.y(),
                availableGeometry.width(),availableGeometry.height(),
                );
    }
*/
    connect(desktop,SIGNAL(workAreaResized(int)), this,SLOT(workAreaResized(int))); //FIXME: no this slot activation under Qt5
    // try, so
    QScreen *scrn = qApp->screens().at(screenNum); //qApp->primaryScreen();
    connect(scrn,SIGNAL(virtualGeometryChanged(const QRect &)), this,SLOT(workAreaResized(const QRect &)));
    connect(desktop,SIGNAL(screenCountChanged(int)),            this,SLOT(screenCountChanged(int)));
    connect(desktop,SIGNAL(primaryScreenChanged()),             this,SLOT(primaryScreenChanged()));

    setupKeyboard();

    //if (!isAlone())
    {
#ifdef USE_DOCK
        dock = new KbdDock(this,windowFlags()); //A.Volkov suspects this in uncompressed resize events
#endif
        menu  = createMenu();

        if (QSystemTrayIcon::isSystemTrayAvailable())
        {
            tray = new KbdTray(this);
            if (cfg.value("buttonMenu", true).toBool()) tray->setContextMenu(menu);
            tray->show();
        }
    }

    if (menu && buttonMenu) buttonMenu->setMenu(menu);

    restoreSettings();
    calculateSize();

    lockWidget(strut);

    connect(qApp, SIGNAL(aboutToQuit()), this, SLOT(close()));

#ifdef QAccessibilityClient_FOUND
    if (accessibilityRegistry)
    {
        showHideTimer = new QTimer(this);
        showHideTimer->setSingleShot(true);
        connect(showHideTimer, SIGNAL(timeout()), this, SLOT(showHideTimeout()));
    }
#endif

#ifdef WITH_SOLID
   isKbdAttached = findKbdAttached();

   DeviceNotifier *devNotifier = DeviceNotifier::instance();
   if (devNotifier) 
   {
     connect(devNotifier,SIGNAL(deviceAdded(const QString&)),  this,SLOT(deviceChangedSlot(const QString&)));
     connect(devNotifier,SIGNAL(deviceRemoved(const QString&)),this,SLOT(deviceChangedSlot(const QString&)));
   }
#endif
}

MainWidget::~MainWidget()
{
    if (tray) { delete tray; tray=NULL; }
    if (dock) { delete dock; dock=NULL; }
#ifdef WITH_SONNET
    if (speller)        { delete speller;       speller=NULL; }
    //if (guesser)      { delete guesser;       guesser=NULL; }
    if (currentString)  { delete currentString; currentString=NULL; }
#endif
}

void  MainWidget::setArgGeometry(const QString &gm)
{
    if (strut != MainWidget::StrutNone && 
        strut != MainWidget::StrutBottom) 
    {
        fprintf(stderr,"Can't apply geom arg if strut arg is set\n");
        return;
    }

    int x=-INT_MAX,y=-INT_MAX;
    unsigned int w=0,h=0;

    int mask = XParseGeometry(gm.toLocal8Bit().data(),&x,&y,&w,&h);
    const auto dpr = devicePixelRatioF();
    x = qRound(x / dpr);
    y = qRound(y / dpr);
    w = qRound(w / dpr);
    h = qRound(h / dpr);
    /*if (!(mask&WidthValue))  w=0;
    if (!(mask&HeightValue)) h=0;
    if (!(mask&XValue))      x=-1;
    if (!(mask&YValue))      y=-1;
    if (mask&XNegative)    ;
    if (mask&YNegative)    ;*/

    if (x==-INT_MAX) x=0;
    if (y==-INT_MAX) y=0;
    if (w==0) w = (screenGeometry.width()>screenGeometry.height()) ? screenGeometry.width()/2  : screenGeometry.width();
    if (h==0) h = (screenGeometry.width()>screenGeometry.height()) ? screenGeometry.height()/4 : screenGeometry.height()/6;

    if (x<0 || mask&XNegative) x = screenGeometry.width()  + x - w;
    if (y<0 || mask&YNegative) y = screenGeometry.height() + y - h;

    //fprintf(stderr,"%s: x=%d y=%d w=%d h=%d\n",gm.toLocal8Bit().data(),x,y,w,h);
    setGeometry(QRect(x,y,w,h));

    geom = QRect(x,y,w,h);
    if (/*!isAlone() &&*/ strut == MainWidget::StrutBottom) setStrutGeometry();
}

#ifdef WITH_SOLID
bool MainWidget::findKbdAttached()
{
    QList<Device> devList = Device::listFromType(DeviceInterface::Keyboard);

    if (getenv("FLY_VKBD_PRINT_KBD_ITEMS")) 
    {
        int kc=0;
        foreach (Device dev, devList)
        {
            //there are 2 solid kbd without vendor/product strings: PS2 and virtual keyboard - skip them
            fprintf(stderr,"fly-vkbd start: attached keyboard %d of %d: vendor='%s' product='%s' desc='%s'\n",kc,devList.count(),
                        dev.vendor().toLocal8Bit().data(),dev.product().toLocal8Bit().data(),dev.description().toLocal8Bit().data());
            kc++;
        }
    }
    //there is one ACPI Virtual Keyboard always, one more - real attached
    return (devList.count()>1);
}
void MainWidget::deviceChangedSlot(const QString& udi)
{
  if (getenv("FLY_VKBD_PRINT_KBD_ITEMS")) fprintf(stderr,"fly-vkbd: device set changed, look for keyboard(s)\n");
  isKbdAttached = findKbdAttached();
}
#endif

void MainWidget::restoreSettings()
{
    QSettings cfg;

#ifdef QAccessibilityClient_FOUND
    if (cfg.value("accessibility",true).toBool() &&
		!qgetenv("QT_ACCESSIBILITY").isEmpty()) 
    {
        //if not set, then enable accessibility
        //if (qgetenv("QT_LINUX_ACCESSIBILITY_ALWAYS_ON").isEmpty())
            QProcess::execute("fly-set-a11y.sh");

        //subscribe to focus events from registry
        accessibilityRegistry=new QAccessibleClient::Registry(this);
        accessibilityRegistry->subscribeEventListeners(QAccessibleClient::Registry::Focus);
        connect(accessibilityRegistry, SIGNAL(focusChanged(QAccessibleClient::AccessibleObject)),
                this,   SLOT(focusChanged(QAccessibleClient::AccessibleObject)));

        focusRoles = cfg.value("focusRoles",QStringList()).toStringList();
        if (focusRoles.isEmpty())
        {
            focusRoles << "text"      //most common
                       << "entry"     //firefox
                       << "spin box"
                       << "spin button"  //gtk
                       << "combo box"
                       << "frame"      //asked by V. Silantiev to be blacklisted for FF, but libreoffice needs it
                       << "terminal"   //konsole, new fly-term
                       << "paragraph"; //libreoffice specific
            cfg.setValue("focusRoles",focusRoles);
        }
        focusNoRoles = cfg.value("focusNoRoles",QStringList()).toStringList();
        if (focusNoRoles.isEmpty())
        {
            focusNoRoles << "push button"
                         << "document web" //asked by V. Silantiev for FF
                         //<< "frame"      //asked by V. Silantiev for FF
                         << "link";        //asked by V. Silantiev for FF
            cfg.setValue("focusNoRoles",focusNoRoles);
        }

        autoShowDelay = cfg.value("autoShowDelay",500).toInt();
    }
    else
    {
        autoShowDelay = 0;
        accessibilityRegistry=NULL;
    }
#endif
    bool locked         = cfg.value("locked", false).toBool();
    bool autoResize     = cfg.value("autoresfont", true).toBool();
    QFont font          = cfg.value("font", QFont()).value<QFont>();
    bool extentVisible  = cfg.value("extentVisible", false).toBool();
    bool showDock       = cfg.value("showdock", false).toBool();
    processRightButton  = cfg.value("processRightButton", false).toBool();
    fixedRightShift     = cfg.value("fixedRightShift", true).toBool();

    unsigned int dfltAutoRepeatDelay=1000, dfltAutoRepeatRate=50;
    /*QStyleHints *styleHints = qApp->styleHints();
    if (styleHints)
    { 
        //400, 30 are default in qt5 hints
        dfltAutoRepeatDelay = styleHints->keyboardInputInterval();
        dfltAutoRepeatRate  = styleHints->keyboardAutoRepeatRate();
    } else*/
    {
      //660, 40 are defaults in xkb, but xkb API requires device id in XkbGetAutoRepeatRate()
      XkbGetAutoRepeatRate(QX11Info::display(), XkbUseCoreKbd, &dfltAutoRepeatDelay,&dfltAutoRepeatRate);
    }
    vkbdAutoRepeatDelay = cfg.value("autoRepeatDelay", dfltAutoRepeatDelay).toInt();
    vkbdAutoRepeatRate  = cfg.value("autoRepeatRate",  dfltAutoRepeatRate).toInt();


    setBackgroundColor(bgColor);

    if (cssStyle.contains("%1"))
    {
        //get colors
        if (bgColor.name()=="#000000") bgColor = QColor("#222222");

        QString baseColor  = bgColor.name();
        QString baseColor1 = bgColor.lighter(100).name();
        QString baseColor2 = bgColor.darker(120).name();//"#52A0AF";
        QString baseColor3 = bgColor.darker(150).name();//"#345174";

        QString tbColor1  = bgColor.lighter(150).name();//"#ddd"; //some light gray
        QString tbColor2  = bgColor.lighter(250).name();//"#fff"; //a bit lighter then tbColor1

        qreal R = 0.3  * bgColor.red();
        qreal G = 0.59 * bgColor.green();
        qreal B = 0.11 * bgColor.blue();

        QString fontColor = (R+G+B+1.0) < 128.0 ? "white" : "black"; //"#333"; //inverted from base color
        //------------------

        QString cssStyleFilled;

        if (cssStyle.contains("%32"))
        {
            cssStyleFilled = cssStyle.arg("F1").arg("F2").arg("F3").arg("F4").
                                arg(fontColor).arg(baseColor2). //5,6
                                arg(tbColor1).arg(baseColor). //7,8
                                arg(tbColor2).arg(baseColor).//9,10
                                arg(baseColor2).arg(tbColor1).//11,12
                                arg(baseColor2).arg(tbColor1)//13,14
                                .arg(baseColor2).arg(tbColor1) //15,16
                                .arg(fontColor) //17
                                .arg("F1").arg("F2").arg("F3").arg("F4") //18,19,20,21
                                .arg(fontColor).arg(baseColor2) //22,23
                                .arg(tbColor1).arg(baseColor) //24,25
                                .arg(tbColor2).arg(tbColor1) //26,27
                                .arg(baseColor2).arg(tbColor1) //28,29
                                .arg(baseColor2).arg(tbColor1) //30,31
                                .arg(baseColor); //32
        }
        else if (cssStyle.contains("%23"))
        {
            cssStyleFilled = cssStyle.arg("F1").arg("F2").arg("F3").arg("F4").
                                arg(fontColor).arg(baseColor3). //5,6
                                arg(baseColor1). //7
                                arg(tbColor1). //8
                                arg(tbColor1).//9
                                arg(tbColor1)//10
                                .arg(baseColor2) //11
                                .arg(fontColor)  //12
                                .arg("F1").arg("F2").arg("F3").arg("F4") //13,14,15,16
                                .arg(fontColor).arg(baseColor2) //17,18
                                .arg(baseColor2) //19
                                .arg(tbColor1)//20
                                .arg(baseColor) //21
                                .arg(baseColor) //22
                                .arg(baseColor2);//23
        }
        cssStyle = cssStyleFilled;
    }

	cssStyleOrig = cssStyle; //save initial style to restore if re-called with "-s default"

	//fprintf(stderr,"vkbd css style '%s'\n",cssStyle.toLocal8Bit().data());

	setLocked(locked);
	setAutoResized(autoResize);
	//updateFont(font); was called in setAutoResized

	if (extentVisible) {
		isExtentVisible = false;
		toggleNumericPad();
	}

        strutBottomHeight = cfg.value("strutBottomHeight", "0.33").toDouble();

    if (cfg.contains("opacity"))
    {
    /*
        unsigned int v = cfg.value("opacity",100).toInt()/100.*0XFFFFFFFF;
        XChangeProperty(display, winId(), 
                    XInternAtom(display,"_NET_WM_WINDOW_OPACITY", false), 
                    XA_CARDINAL, 32, PropModeReplace, (unsigned char *)(&v), 1L);
    or in Qt:*/
        setWindowOpacity(cfg.value("opacity",100).toDouble()/100.0);
    }

	if (dock) {
		dock->restoreState();
		dock->setVisible(/* !isAlone() &&*/ showDock);
	}

	restorePosition();
}

//bool MainWidget::isAlone() const { return alone; }

bool MainWidget::autoResized() const
{
	return autoResize;
}

void MainWidget::setAutoResized(bool mode)
{
	autoResize = mode;
	QFont font = QSettings().value("font", QFont()).value<QFont>();
	qDebug() << Q_FUNC_INFO << font.pointSize();
	updateFont(font);
	emit(autoResizeToggled(mode));
}

void MainWidget::restorePosition()
{
	if (/*!isAlone() &&*/ strut) setStrutGeometry();
    else
    {
        //QRect screenGeom = desktop->screenGeometry();
        QDesktopWidget *dw = QApplication::desktop();
        QRect screenGeom = dw->screenGeometry(dw->primaryScreen());

        int dWidth = 0;
        if (isExtentVisible) {
            dWidth = 700;
        } else {
            dWidth = 560;
        }
        int dHeight = VButton::ph;
        QRect dfltGeom(screenGeom.width() - dWidth, screenGeom.height() - dHeight, dWidth, dHeight);

        //if (!isAlone())
        {
            QSettings cfg;
            QRect geom  = cfg.value("geometry", QRect()).toRect();
            if (!geom.isNull() && geom.isValid()) {
                dfltGeom = geom;
            }
        }

        setGeometry(dfltGeom);
        qDebug() << "set geometry to " << dfltGeom;
    }
}

void MainWidget::saveState()
{
	//if (isAlone()) return;

    if (strut) return;

	QSettings cfg;
	if (cfg.isWritable()) {
		//cfg.setValue("visible", isVisible());
		cfg.setValue("geometry", geometry());
		cfg.sync();
	}
}

void MainWidget::calculateSize()
{
	//QRect screenGeom = desktop->screenGeometry();
    QDesktopWidget *dw = QApplication::desktop();
    QRect screenGeom = dw->screenGeometry(dw->primaryScreen());

	int w = qMin(width(),screenGeom.width()); //alex

	if (isExtentVisible) {
		sdxb = w;
		sdxs = w - (w * (140.0 / 700.0));
		VButton::pw = 700;
		setMinimumSize(700 * 2 / 3, VButton::ph * 2 / 3);
	} else {
		sdxs = w;
		sdxb = w + (w * (140.0 / 560.0));
		VButton::pw = 560;
		setMinimumSize(560 * 2 / 3, VButton::ph * 2 / 3);
	}
}

QString MainWidget::setFontInStyle(const QString& cssStyle, const QFont& fnt)
{
	QString res;

	if (cssStyle.contains("F1") && cssStyle.contains("F2") && cssStyle.contains("F3") && cssStyle.contains("F4"))
	{
		res = cssStyle;
		res = res.replace(QString("F1"),fnt.family());
		res = res.replace(QString("F2"),QString::number(fnt.pointSize()));
		res = res.replace(QString("F3"),fnt.styleName());
		res = res.replace(QString("F4"),fontWeightString(fnt.weight()));
	}
	else
	{
		res = cssStyle.arg(fnt.family()).arg(fnt.pointSize()).arg(fnt.styleName()).arg(fontWeightString(fnt.weight()));
		//fprintf(stderr,"\nres=%s\nstyle=%s\n\n",res.toLocal8Bit().data(),cssStyle.toLocal8Bit().data());
	}

	return res;
}

void MainWidget::hideByUserSlot()
{
    visibilityForced=false;
    hide();
}

void MainWidget::closeEvent(QCloseEvent */*e*/)
{
	saveState();
	if (dock) dock->saveState();
}

void MainWidget::resizeEvent(QResizeEvent * e)
{
	if (!resizable) return;

    //fprintf(stderr,"resize event %d %d winhndl=%x\n",e->size().width(),e->size().height(),windowHandle());

	if (e->spontaneous()) return; //qt5 speed up

	setUpdatesEnabled(false); //qt5 speed up

	QSize sz = e->size();

	calculateSize();

	//QRect screenGeom = desktop->screenGeometry();
    QDesktopWidget *dw = QApplication::desktop();
    QRect screenGeom = dw->screenGeometry(dw->primaryScreen());

	if (!strut && sz.width()>screenGeom.width()) //alex
	{
	    sz.setWidth(screenGeom.width());
	    //fprintf(stderr,"fly-vkbd: resizeEvent(): possible infinite loop resizeEvent()->resize()->resizeEvent()... "
		//	   "may be just ignore event by e->ignore() and return - side effects are possible \n");
	    QWidget::resize(sz);
	}

	if (autoResized()) { //qt5 drastic speed up: do it before all other changes in widgets to avoid extra re-creating of gradients
	  double rp = (8.0 / 600.0) * sdxs * dpiMul;
	  QFont fnt = this->font();
	  if (fabs(fnt.pointSizeF()-rp)>1.5) //qt5 speed up
	  {
		fnt.setPointSize/*F*/((int)(rp*0.9));
		setFont(fnt);
		if (!cssStyle.isEmpty())
		{
			QString style = setFontInStyle(cssStyle,fnt); //cssStyle.arg(fnt.family()).arg(fnt.pointSize()).arg(fnt.styleName()).arg(fontWeightString(fnt.weight()));

			//if (bgColor.isValid())   setBackgroundColor(bgColor); //bg must be in style
            if (style!=styleSheet()) setStyleSheet(style);
		}
	  }
	}

	for (int a = 0; a < buttonsList.size(); a++) {
		VButton *v = buttonsList.at(a);
		v->reposition(sz.width(), sz.height());
	}
	for (int a = 0; a < modKeys.size(); a++) {
		VButton *v = modKeys.at(a);
		v->reposition(sz.width(), sz.height());
	}
	for (int a = 0; a < otherKeys.size(); a++) {
		VButton *v = otherKeys.at(a);
		v->reposition(sz.width(), sz.height());
	}
	for (int a = 0; a < numLockKeys.size(); a++) {
		VButton *v = numLockKeys.at(a);
		v->reposition(sz.width(), sz.height());
	}

	//QRect mainKeysRect = (funcRow?esc->geometry().united(rightCtrl->geometry().united(rightShift->geometry())) : buttonsList[0]->geometry().united(/*rightCtrl*/esc->geometry()));
    QRect mainKeysRect = (funcRow? esc->geometry().united(layoutSwitcher->geometry()) : 
                                   buttonsList[0]->geometry().united(layoutSwitcher->geometry()));
    QRect numKeysRect = numLock->geometry().united(ent->geometry());
	QRegion r(mainKeysRect);
	r |= numKeysRect;

	if (quit && quit->isVisibleTo(this))
	{
		r |= (quit->geometry().united(extent->geometry()));
	}

	setIgnoreDragRegion(r);


	int isize = 0, margin=10;
	if (layoutSwitcher) 
	{
		isize = GetStandardIconSize(qMin(layoutSwitcher->width(),layoutSwitcher->height())-margin);
		layoutSwitcher->setIconSize(QSize(isize,isize));
	}
	if (buttonMenu) 
	{
		if (!isize) isize = GetStandardIconSize(qMin(buttonMenu->width(),buttonMenu->height())-margin);
		buttonMenu->setIconSize(QSize(isize,isize));
	}
#ifdef WITH_SONNET
	if (suggestMenu) 
	{
		if (!isize) isize = GetStandardIconSize(qMin(suggestMenu->width(),suggestMenu->height())-margin);
		suggestMenu->setIconSize(QSize(isize,isize));

        if (suggestTable)
        {
          suggestMenu->setEnabled(false);
          delete suggestTable; suggestTable=NULL;
        }
	}
#endif
	if (extent) 
	{
		if (!isize) isize = GetStandardIconSize(qMin(extent->width(),extent->height())-margin);
		extent->setIconSize(QSize(isize,isize));
	}
	if (quit) 
	{
		if (!isize) isize = GetStandardIconSize(qMin(quit->width(),quit->height())-margin);
		quit->setIconSize(QSize(isize,isize));
	}

	setUpdatesEnabled(true); //qt5 speed up
}

void MainWidget::primaryScreenChanged()  //alex
{
    QDesktopWidget *dw = QApplication::desktop();
    screenNum  = dw->primaryScreen();
    QRect scrg = dw->screenGeometry(screenNum);
    screenGeometry = scrg;
    //fprintf(stderr,"fly-vkbd: primary screen geom = %ld %ld %ld %ld strut=%d, screen=%d\n\t",scrg.x(),scrg.y(),scrg.width(),scrg.height(),strut,screenNum);

    if (strut) setStrutGeometry(); //workAreaResized(screenNum);
}

void MainWidget::screenCountChanged(int num) //alex
{
    Q_UNUSED(num);
    QDesktopWidget *dw = QApplication::desktop();
    screenNum  = dw->primaryScreen();
    QRect scrg = dw->screenGeometry(screenNum);
    screenGeometry = scrg;
    //fprintf(stderr,"fly-vkbd: count changed, primary screen geom = %ld %ld %ld %ld strut=%d screen=%d\n\t",scrg.x(),scrg.y(),scrg.width(),scrg.height(),strut,screenNum);

    if (strut) setStrutGeometry(); //workAreaResized(screenNum);
}

void MainWidget::workAreaResized(const QRect & r) //alex
{
    Q_UNUSED(r);
    QDesktopWidget *dw = QApplication::desktop();
    screenNum  = dw->primaryScreen();
    //QRect scrg = dw->screenGeometry(screenNum);
    //fprintf(stderr,"work area resized to rect = %d %d %d %d screen=%d\n\t",r.x(),r.y(),r.width(),r.height(),screenNum);

    workAreaResized(screenNum); //if (strut) setStrutGeometry();
}

void MainWidget::workAreaResized(int scr)
{
    QDesktopWidget *dw = QApplication::desktop();
    screenNum  = dw->primaryScreen();
    QRect scrg = dw->screenGeometry(screenNum);
    //fprintf(stderr,"work area of screen=%d (our is %d) resized to %d %d %d %d old=%d %d %d %d\n",
      //      scr,screenNum,scrg.x(),scrg.y(),scrg.width(),scrg.height(),
        //    screenGeometry.x(),screenGeometry.y(),screenGeometry.width(),screenGeometry.height());

	if (screenNum!=scr) return; //not our screen?

	if (screenGeometry==scrg) return; //no real changes (often generated may be due to our strut property)
	screenGeometry = scrg;

	if (strut) setStrutGeometry();
}

void MainWidget::updateFont(QFont &fnt)
{

//  fnt.setWeight(QFont::Bold);
	if (autoResized() && sdxs>0) {
        double rp = (8.0 / 600.0) * /*width()*/sdxs * dpiMul;
        fnt.setPointSize/*F*/((int)(rp*0.9));
	}
	setFont(fnt);
	if (!cssStyle.isEmpty())
	{
	  //QString family = fnt.family();
	  //foreach (const QChar& c, family) fprintf(stderr, "%d ", c.unicode()); fprintf(stderr, "\n");

      QString style = setFontInStyle(cssStyle,fnt);//cssStyle.arg(fnt.family()).arg(fnt.pointSize()).arg(fnt.styleName()).arg(fontWeightString(fnt.weight()));

      //fprintf(stderr,"updateFont: style=%s\n         cssStyte=%s\n",style.toLocal8Bit().data(),cssStyle.toLocal8Bit().data());
      
	  //if (bgColor.isValid())   setBackgroundColor(bgColor); //bg must be in style
      if (style!=styleSheet()) setStyleSheet(style);
	}
}
void MainWidget::toggleNumericPad()
{
	QSize sizeOld = size();
	resizable = false;
	QSettings cfg;
	cfg.setValue("extentVisible", !isExtentVisible);
	cfg.sync();

    //QRect screenGeom = desktop->screenGeometry();
    QDesktopWidget *dw = QApplication::desktop();
    QRect screenGeom = dw->screenGeometry(dw->primaryScreen());

	if (isExtentVisible) {
		isExtentVisible = false;
		if (strut) calculateSize();
		QWidget::resize((int)sdxs, height());
		extent->setIcon(QIcon::fromTheme("view-right-new"));
		///extent->setToolTip(tr("Show Numeric Keypad"));
	} else {
		isExtentVisible = true;
		if (strut) calculateSize();
		QWidget::resize((int)sdxb, height());
		extent->setIcon(QIcon::fromTheme("view-right-close"));
		///extent->setToolTip(tr("Hide Numeric Keypad"));

		// Make sure the numpad doesn't go beyond the screen edge
		int screenRight = screenGeom.x() + screenGeom.width();
		int widgetRight = x() + width();
		int overlap = widgetRight - screenRight;
		if (!strut && overlap > 0) {
			move((screenRight - width())>0?(screenRight - width()):0, y());
		}
	}
	resizable = true;

	if (strut || width()>screenGeom.width()) //alex
	{
	    QResizeEvent re(size(),sizeOld);
	    resizeEvent(&re);
	    repaint();
	}

#ifdef WITH_SONNET
    resetSuggestedMenu();
#endif
}

void MainWidget::chooseFont()
{
	// Hide the widget before showing the dialog to avoid overlapping it
	bool toShow = false;
	if (isVisible()) {
		hide();
		toShow = true;
	}

	QFont font = this->font();
	bool ok;
	font = QFontDialog::getFont(&ok, font, this);
	if (ok)
	{
		QSettings cfg;
		cfg.setValue("font", font);
		cfg.sync();
		updateFont(font);
	}

	if (toShow) {
		show();
	}
}
/*
void MainWidget::chooseBackgroundColor()
{
	bool toShow = false;
	if (isVisible()) {
		hide();
		toShow = true;
	}

	QColor color = palette().color(backgroundRole());
	QColor newColor = QColorDialog::getColor(color, this, tr("Choose keyboard background color"));
	if (newColor.isValid())
	{
		QSettings cfg;
		cfg.setValue("bg", newColor.name());
		cfg.sync();

        bgColor = newColor;
		setBackgroundColor(bgColor); //newColor
	}

	if (toShow) {
		show();
	}
}
*/
void MainWidget::lockWidget(bool lock)
{
	setLocked(lock);
	QSettings cfg;
	cfg.setValue("locked", lock);
	cfg.sync();

	repaint(); //alex
}

void MainWidget::showDock(bool showit)
{
	if (!dock) dock = new KbdDock(this,windowFlags());
	dock->setVisible(showit);
	QSettings cfg;
	cfg.setValue("showdock", showit);
	cfg.sync();
}

void MainWidget::autoResizeFont(bool autores)
{
	setAutoResized(autores);
	QSettings cfg;
	cfg.setValue("autoresfont", autores);
	cfg.sync();
}

void MainWidget::autoShowSlot(bool val)
{
    autoShow=val;
	QSettings cfg;
    cfg.setValue(flySession+"/autoShow",autoShow);
	cfg.sync();

    if (autoShow /*&& qgetenv("QT_LINUX_ACCESSIBILITY_ALWAYS_ON").isEmpty()*/)
        QProcess::execute("fly-set-a11y.sh");
}

void MainWidget::funcRowSlot(bool val)
{
    funcRow=val;
	QSettings cfg;
	cfg.setValue("funcRow", funcRow);
	cfg.sync();

    //if (QMessageBox::question(this,tr("Attention"),tr("This setting will apply after this program restart. Restart now?"))!=QMessageBox::Yes) return;

    //save and hide self
    saveState();
    hide();
    if (tray) tray->setVisible(false);

    //initiate update
    QStringList args; args << "--update";
    QProcess::startDetached(qApp->applicationFilePath(),args);
}

void MainWidget::autoRepeatSlot(bool val)
{
    autoRepeated=val;
	QSettings cfg;
	cfg.setValue("autoRepeated", autoRepeated);
	cfg.sync();
}

void MainWidget::toggleNumlock()
{
	bool p = numLock->isChecked();
	for (int a = 0; a < numLockKeys.size(); a++) {
		NumpadVButton *v = numLockKeys[a];
		v->numLockPressed(p);
	}

	if (rightShift && fixedRightShift) //alex
	{
		rightShift->setChecked(false); 
		resetKeysCap();
	}
}

void MainWidget::toggleCaps()
{
	if (modTimer) modTimer->stop();
	sendKey(caps->getKeyCode());
	if (modTimer) modTimer->start();

	if (rightShift && fixedRightShift) rightShift->setChecked(false); //alex

	resetKeysCap();
}

void MainWidget::resetKeysCap()
{
	qDebug() << Q_FUNC_INFO;
	for (int a = 0; a < buttonsList.size(); a++) {
		VButton *v = buttonsList[a];
		v->setKeyCap(leftShift->isChecked() || rightShift->isChecked(), caps->isChecked());
	}
}

void MainWidget::keyPress(unsigned int a)
{
	// Prevent modKeys state from changing state three times in a row
	// because the modTimer timed out when a modKey was clicked
	if (modTimer) modTimer->stop();
	sendKey(a);
	if (modTimer) modTimer->start();

	for (int i = 0; i < modKeys.size(); i++) {
		VButton *mod = (VButton*) modKeys[i];
		if (mod==rightShift && fixedRightShift) continue; //alex
		mod->setChecked(false);
	}

	resetKeysCap();
}

void MainWidget::sendKey(unsigned int keycode)
{/*
	Window currentFocus;
	int revertTo;

	XGetInputFocus(display, &currentFocus, &revertTo);
*/
	for (int a = 0; a < modKeys.size(); a++) {
		VButton *mod = (VButton*) modKeys[a];
		if (mod->isChecked()) {
			XTestFakeKeyEvent(display, mod->getKeyCode(), true, 0);
		}
	}

	XTestFakeKeyEvent(display, keycode, true, 1);
	XTestFakeKeyEvent(display, keycode, false,2);

	for (int a = 0; a < modKeys.size(); a++) {
		VButton *mod = (VButton*) modKeys[a];
		if (mod->isChecked()) {
			XTestFakeKeyEvent(display, mod->getKeyCode(), false, 2);
		}
	}

#ifdef WITH_SONNET
    updateSuggestedMenu();
#endif

	XFlush(display);
}

#ifdef WITH_SONNET

class MyItemDelegate : public QStyledItemDelegate
{
  int height;

public:
  MyItemDelegate(int h, QObject* parent = nullptr) : QStyledItemDelegate(parent)
  {
    height=h;
  }
     
  QSize sizeHint(const QStyleOptionViewItem & option, const QModelIndex & index) const override
  {
    return QSize(QStyledItemDelegate::sizeHint(option,index).width(), height);
  }
};

void MainWidget::updateSuggestedMenu()
{
    if (speller && suggestMenu && suggestMenu->isVisible() && currentString)
    {
        if (currentString->isEmpty())
        {
            suggestMenu->setEnabled(false);
            if (suggestTable) { delete suggestTable; suggestTable=NULL; }
        }
        else
        {
            QString slangOld = speller->language();
            QString slangNew = slangOld;

            //simple check if Latin or not
            ushort ucs = currentString->at(0).unicode();
            if (KeySym2Ucs::isUCSLatin(ucs)) 
                 { if (slangOld!="en") slangNew="en"; }
            else { if (slangOld!="ru") slangNew="ru"; }

            if (slangOld!=slangNew) speller->setLanguage(slangNew);

            if (suggestTable) suggestTable->clear(); 
            else
            {
                int x = 0, y = 0;
                if (funcRow) 
                {
                    x = esc->pos().x();
                    y = esc->pos().y();
                }
                else
                {
                    x = buttonsList[0]->pos().x();
                    y = buttonsList[0]->pos().y();
                }


                //fprintf(stderr,"for lang=%s suggest=",slangNew.toLocal8Bit().data());
                suggestTable = new MyListWidget(
                    QSize( bksp->x()- x - hKeySpace, buttonsList[0]->size().height() ),
                    QSize( bksp->x()- x - hKeySpace, height() - y - vKeySpace),
                    QPoint(x,y),
                    this);
                suggestTable->setWrapping(true);
                suggestTable->setFlow(QListView::LeftToRight);
                suggestTable->setFont(buttonsList[0]->font());

                connect(suggestTable,SIGNAL(itemClicked(QListWidgetItem *)),this,SLOT(suggested(QListWidgetItem *)));

                //to set custom item height
                suggestTable->setItemDelegate(new MyItemDelegate(buttonsList[0]->size().height(),this));

                /* or
                suggestTable = new QTableWidget(1,1,this);
                suggestTable->setSelectionMode(QTableWidget::NoSelection);
                suggestTable->horizontalHeader()->hide();
                suggestTable->verticalHeader()->hide();
                suggestTable->setRowHeight(0,buttonsList[0]->size().height());
                connect(suggestTable,SIGNAL(cellClicked(int,int)),this,SLOT(suggested(int,int)));
                */

                //setup scrolling---------------
                //QList<const QTouchDevice *>tdl = QTouchDevice::devices();
                //if (tdl.count())
                {
                    Qt::ScrollBarPolicy scrollBarPolicy = Qt::ScrollBarAsNeeded;
                    //if (cms::crossplatform::isTouchDevice() && cms::crossplatform::isMobileDevice())
                    {
                        // QAbstractItemView::ScrollPerPixel results in smooth scrolling compared to ScrollPerItem
                        suggestTable->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
                        suggestTable->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
                        scrollBarPolicy =  Qt::ScrollBarAlwaysOff;
                    }
                    suggestTable->setHorizontalScrollBarPolicy(scrollBarPolicy);
                    suggestTable->setVerticalScrollBarPolicy(scrollBarPolicy);

                    QScroller::grabGesture(suggestTable,QScroller::LeftMouseButtonGesture);

                    QScrollerProperties properties = QScroller::scroller(suggestTable)->scrollerProperties();
                    QVariant overshootPolicy = QVariant::fromValue<QScrollerProperties::OvershootPolicy>
                                                (QScrollerProperties::OvershootAlwaysOff);
                    properties.setScrollMetric(QScrollerProperties::HorizontalOvershootPolicy,overshootPolicy);
                    properties.setScrollMetric(QScrollerProperties::VerticalOvershootPolicy,overshootPolicy);
                    QScroller::scroller(suggestTable)->setScrollerProperties(properties);
                }
                //setup scrolling end---------------
            }

            QStringList slist = speller->suggest(*currentString);

            if (slist.indexOf(*currentString)<0) //add current word too
            {
               /*QTableWidgetItem *item = new QTableWidgetItem(*currentString);
               item->setIcon(newIcon);
               suggestTable->setColumnCount(j+1);
               suggestTable->setItem(0,j++,item);*/
               QListWidgetItem * item = new QListWidgetItem(*currentString);
               item->setIcon(newIcon);
               suggestTable->addItem(item);

               //fprintf(stderr,"input=%s", currentString->toLocal8Bit().constData());
            }

            for (int i = 0; i < slist.size(); i++) 
            {
               /*QTableWidgetItem *item = new QTableWidgetItem(slist.at(i));
               suggestTable->setColumnCount(j+1);
               suggestTable->setItem(0,j++,item);*/
               QListWidgetItem * item = new QListWidgetItem(slist.at(i));
               suggestTable->addItem(item);

               //fprintf(stderr,"input=%s", currentString->toLocal8Bit().constData());
               //fprintf(stderr,"%s,", slist.at(i).toLocal8Bit().constData());
            }


            if (!speller->isCorrect(*currentString))
            {
                //guess of input lang
                //slangNew = guesser->identify(*currentString);

                QString transString;
                if (slangNew=="en") { slangNew="ru"; transString = lat2rus(*currentString); }
                else                { slangNew="en"; transString = rus2lat(*currentString); }
                speller->setLanguage(slangNew);

                QStringList olist = speller->suggest(transString);

                for (int i = 0; i < olist.size(); i++) 
                {
                   //fprintf(stderr,"%s,", olist.at(i).toLocal8Bit().constData());
                   QListWidgetItem * item = new QListWidgetItem(olist.at(i));
                   suggestTable->addItem(item);
                   /*
                   QTableWidgetItem *item = new QTableWidgetItem(olist.at(i));
                   suggestTable->setColumnCount(j+1);
                   suggestTable->setItem(0,j++,item);
                   */
                }
            }


            if (suggestTable && suggestTable->count()) 
            {
                suggestMenu->setEnabled(true);
                suggestTable->show(); //auto show suggestions
            }


            //fprintf(stderr,"\n");

            if (slangOld!=slangNew) speller->setLanguage(slangOld);
        }
    }
}

void MainWidget::resetSuggestedMenu()
{
    if (currentString) currentString->clear();

    if (suggestMenu) 
    {
        suggestMenu->setEnabled(false);
        if (suggestTable) 
        { 
            //crash if so
            //delete suggestTable; suggestTable=NULL; 
            suggestTable->clear();
            suggestTable->hide(); 
        }
    }
    fromClipboard = false;
}

void MainWidget::showSuggestTable(bool)
{
    if (!suggestTable) return;

    if (suggestTable->isVisible()) 
         suggestTable->hide();
    else suggestTable->show();
}

//void MainWidget::suggested(int row,int col)
void MainWidget::suggested(QListWidgetItem *q)
{
    //QTableWidgetItem *q = suggestTable->item(row,col);
    if (!q) return;

    QString txt = q->text();
    int szOld = fromClipboard? 1 : currentString->size();
    int szNew = txt.size();

    //own input word is selected add it to personal
    //fprintf(stderr,"curr='%s' selected='%s'\n",currentString->toLocal8Bit().data(),txt.toLocal8Bit().data());
    if (currentString && txt == *currentString)
    {
        if (q && !q->icon().isNull())
        {
                //fprintf(stderr,"curr='%s' selected='%s' acted='%s'\n",currentString->toLocal8Bit().data(),txt.toLocal8Bit().data(),act0->text().toLocal8Bit().data());
                if (q->text() == txt) 
                {
                    speller->addToPersonal(txt);
                    //fprintf(stderr,"%s added to personal\n",txt.toLocal8Bit().data());
                }
        }
    }

    //fprintf(stderr,"select=%s, send %d backspaces and add %d new chars",txt.toLocal8Bit().data(),szOld,szNew);

    //clear original text
    KeyCode kshift = XKeysymToKeycode(display,XK_Shift_L);
    KeyCode kc = XKeysymToKeycode(display, fromClipboard ? XK_Delete : XK_BackSpace);
    for (int i=0;i<szOld;i++)
    {
        XTestFakeKeyEvent(display, kc, true, 1);
        XTestFakeKeyEvent(display, kc, false,2);
    }

    //TODO: check Latin or not and switch group for a while!
    int oldLayout=-1, newLayout;
    ushort ucs = txt.at(0).unicode();
    if (KeySym2Ucs::isUCSLatin(ucs)) 
         newLayout = changeLayoutTo("us","en", &oldLayout);
    else newLayout = changeLayoutTo("ru","cyr",&oldLayout);

    //send suggested text
    for (int i=0;i<szNew;i++)
    {
        const QChar ch = txt.at(i);
        KeySym ks = KeySym2Ucs::ucs2keysym((long)ch.unicode());

        bool needShift=!(ch.isLower());

        kc = XKeysymToKeycode(display,ks);

        if (needShift) XTestFakeKeyEvent(display, kshift, true, 1);

        XTestFakeKeyEvent(display, kc, true, 1);
        XTestFakeKeyEvent(display, kc, false,2);

        if (needShift) XTestFakeKeyEvent(display, kshift, false,2);

        //fprintf(stderr,"%x->%x->%x ",(long)ch.unicode(),ks,kc);
    }


    //restore layout if changed
    if (newLayout!=oldLayout && oldLayout>=0) 
      XkbLockGroup(QX11Info::display(),XkbUseCoreKbd,oldLayout);

    //fprintf(stderr,"\n");
	XFlush(display);

    resetSuggestedMenu();
}

void MainWidget::checkClipboard(QClipboard::Mode mode)
{
    QString txt = clipboard->text(mode);

    if (txt.isEmpty()) return;

    int sz = txt.length();
    for (int i=0;i<sz;i++) { if (!txt[i].isLetter()) return; }

    *currentString = txt;
    fromClipboard=true;

    updateSuggestedMenu();
}


void MainWidget::useSuggestSlot(bool val)
{
    useSonnet=val;
	QSettings cfg;
	cfg.setValue("useSonnet", val);
	cfg.sync();

    //if (QMessageBox::question(this,tr("Attention"),tr("This setting will apply after this program restart. Restart now?"))!=QMessageBox::Yes) return;

    //save and hide self
    saveState();
    hide();
    if (tray) tray->setVisible(false);

    //initiate update
    QStringList args; args << "--update";
    QProcess::startDetached(qApp->applicationFilePath(),args);
}
#endif


bool MainWidget::keyState(int iKey)
{
	int          iKeyMask = 0;
	Window       wDummy1, wDummy2;
	int          iDummy3, iDummy4, iDummy5, iDummy6;
	unsigned int iMask;
	XModifierKeymap* map = XGetModifierMapping(display);
	KeyCode keyCode = XKeysymToKeycode(display, iKey);
	if (keyCode == NoSymbol) return false;
	for (int i = 0; i < 8; ++i) {
		if (map->modifiermap[map->max_keypermod * i] == keyCode) {
			iKeyMask = 1 << i;
		}
	}
	XQueryPointer(display, DefaultRootWindow(display), &wDummy1, &wDummy2, &iDummy3, &iDummy4, &iDummy5, &iDummy6, &iMask);
	XFreeModifiermap(map);
	return (iMask & iKeyMask) != 0;
}

void MainWidget::queryModState()
{
	bool capsState = keyState(XK_Caps_Lock);
	bool numLockState = keyState(XK_Num_Lock);

	if (capsState != caps->isChecked()) {
		caps->setChecked(capsState);
		resetKeysCap();
	}
	if (numLockState != numLock->isChecked()) {
		numLock->setChecked(numLockState);
		toggleNumlock();
	}
}

void MainWidget::setupText(VButton& v, int group)
{
	KeyCode keycode = v.getKeyCode();

	//KeySym keysym = XKeycodeToKeysym(display, keycode, 0);
	//KeySym keysymShift = XKeycodeToKeysym(display, keycode, 1);
	KeySym keysym = keyCodeToKeySym(keycode, group, 0);
	long ret = KeySym2Ucs::keysym2ucs(keysym);
	if (ret == -1)
		ret = KeySym2Ucs::keysym2ucs(keyCodeToKeySym(keycode, 0, 0));

	QString buttonText(QChar((uint)ret));

	// To display an ampersand, the text has to be set to "&&" because
	// "&" is used to create shortcut keys
	if (buttonText == "&") {
		v.setText("&&");
	} else {
		v.setText(buttonText);
	}

	KeySym keysymShift = keyCodeToKeySym(keycode, group, ShiftMask);
	long shiftRet = KeySym2Ucs::keysym2ucs(keysymShift);
	if (shiftRet == -1)
		shiftRet = KeySym2Ucs::keysym2ucs(keyCodeToKeySym(keycode, 0, ShiftMask));
	QChar c((uint)shiftRet);
	if (c == '&') {
		v.setShiftText("&&");
	} else {
		v.setShiftText(c);
	}
	v.setKeyCap(leftShift->isChecked() || rightShift->isChecked(), caps->isChecked());
}

static int emblemSize(int in_size)
{
    return in_size/2;
}

void MainWidget::mappingNotify(/*XMappingEvent* */int group)
{
    qDebug() << Q_FUNC_INFO;
    for (int a = 0; a < buttonsList.size(); a++) {
        VButton *v = (VButton *)buttonsList[a];
        setupText(*v, group);
    }

    QString layout = getLayouts().value(group);
    layoutSwitcher->setVisible(!layout.isEmpty());

    if (!layout.isEmpty())
    {
        QIcon layoutIcon = iconForLayout(layout);
        if (!layoutIcon.isNull())
        {
            layoutSwitcher->setIcon(layoutIcon);

            if (tray) //change tray icon according to layout: compose main icon with layout icon
            {
                QIcon trayIcon = FlyIcon::withOverlay(tray->icon(), Qt::BottomRightCorner, layoutIcon, emblemSize);
                tray->setIcon(trayIcon);
            }
        }
        else layoutSwitcher->setText(layout);
    }
    else     layoutSwitcher->setText(QString());

#ifdef WITH_SONNET
    if (currentString)
    {
        currentString->clear();

        suggestMenu->setEnabled(false);
        if (suggestTable) { delete suggestTable; suggestTable=NULL; }
    }
    if (speller)
    {
        QString newLang = layout;
        if (newLang.compare(QString("us"),Qt::CaseInsensitive)==0 ) newLang="en";
        //fprintf(stderr,"speller set default lang=%s\n",newLang.toLocal8Bit().data());
        speller->setLanguage(newLang);
        //fprintf(stderr,"speller new default lang=%s\n",speller->language().toLocal8Bit().data());
    }
#endif
}

void MainWidget::setBackgroundColor(const QColor & color)
{
    if (!color.isValid()) return;

	QPalette backgroundColor;
	backgroundColor.setColor(backgroundRole(), color);
	setPalette(backgroundColor);
}

void MainWidget::setupKeyboard()
{
	QSettings settings;
/*
    QString bgColorStr = settings.value("bg", "").toString();
	if (!bgColorStr.isEmpty()) bgColor = QColor(bgColorStr);	
	setBackgroundColor(bgColor.isValid()?bgColor:QColor("#6496aa"));
*/
	funcRow = settings.value("funcRow", false).toBool();
	if (funcRow) VButton::ph=205;
	else         VButton::ph=170;

	//setFocusPolicy(Qt::NoFocus); moved to constructor

	isExtentVisible = false;

	int x = stx;
	int y = sty;

	if (funcRow) 
	{
		esc = new VButton(this);
		esc->setKeyCode(9);
		esc->resize(40, esc->height());
		esc->move(x, y);
		//: button
		esc->setText(tr("Esc"));
		esc->res();
		otherKeys.append(esc);
		connect(esc, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

		qDebug() << Q_FUNC_INFO << "space after esc:" << ((30 + hKeySpace)*2 - esc->width()) / 2;
		x += esc->width() + 16;

		for (int a = 0; a < 4; a++) {
			VButton *f = new VButton(this);
			f->setKeyCode(67 + a);
			f->setText('F' + QString("%1").arg(a + 1));
			f->move(x, y);
			x += f->width() + hKeySpace;
			f->res();
			otherKeys.append(f);
			connect(f, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
		}

		x += 16;
		for (int a = 0; a < 4; a++) {
			VButton *f = new VButton(this);
			f->setKeyCode(71 + a);
			f->setText('F' + QString("%1").arg(a + 5));
			f->move(x, y);
			x += f->width() + hKeySpace;
			f->res();
			otherKeys.append(f);
			connect(f, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
		}

		x += 16;
		for (int a = 0; a < 4; a++) {
			VButton *f = new VButton(this);
			f->setKeyCode(75 + a);
			if (a > 1) f->setKeyCode(93 + a);
			f->setText('F' + QString("%1").arg(a + 9));
			f->move(x, y);
			x += f->width() + hKeySpace;
			f->res();
			otherKeys.append(f);
			connect(f, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
		}
	}

	//ROW 1

	x = stx;
	if (funcRow) y += esc->height() + vKeySpace;

	//QString k1= "`1234567890-=";
	//QString k1s = "~!@#$%^&*()_+";
	unsigned int kc1[ROWLENGTH1] = {49, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21};

	for (int a = 0; a < ROWLENGTH1; a++) {
		VButton *v = new VButton(this);
		v->setKeyCode(kc1[a]);
		v->move(x, y);
		x += v->width() + hKeySpace;
		connect(v, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
		buttonsList.append(v);
		v->res();
		//caps_buttonsList.append ( v );
	}

	bksp = new VButton(this);
	bksp->setKeyCode(22);
	bksp->move(x, y);
	bksp->resize(52, bksp->height());
	//: button Backspace key
	bksp->setText(tr("<--"));
	bksp->res();
	otherKeys.append(bksp);
	connect(bksp, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	//ROW 2

	x = stx;
	y += bksp->height() + vKeySpace;

	VButton *tab = new VButton(this);
	tab->setKeyCode(23);
	tab->move(x, y);
	tab->resize(52, tab->height());
	//: button Tabulation key
	tab->setText(tr("Tab"));
	tab->res();
	otherKeys.append(tab);
	connect(tab, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	x += tab->width() + hKeySpace;

	//QString k2= "qwertyuiop";
	//QString k2s = "QWERTYUIOP";
	unsigned int kc2[ROWLENGTH2] = {24, 25, 26, 27, 28, 29, 30, 31, 32, 33};

	for (int a = 0; a < ROWLENGTH2; a++) {
		VButton *v = new VButton(this);
		v->setKeyCode(kc2[a]);
		//v->setText ( k2.mid ( a,1 ) );
		//v->setShiftText ( k2s.mid ( a,1 ) );
		v->move(x, y);
		x += v->width() + hKeySpace;
		v->res();
		connect(v, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
		buttonsList.append(v);
	}

	VButton *lbr = new VButton(this);
	lbr->setKeyCode(34);
	lbr->move(x, y);
	x += lbr->width() + hKeySpace;
	//lbr->setText ( "[" );
	//lbr->setShiftText ( "{" );
	lbr->res();
	connect(lbr, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(lbr);

	VButton *rbr = new VButton(this);
	rbr->setKeyCode(35);
	rbr->move(x, y);
	x += rbr->width() + hKeySpace;
	//rbr->setText ( "]" );
	//rbr->setShiftText ( "}" );
	rbr->res();
	connect(rbr, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(rbr);

	VButton *bksl = new VButton(this);
	bksl->setKeyCode(51);
	bksl->move(x, y);
	//bksl->setText ( "\\" );
	//bksl->setShiftText ( "|" );
	bksl->res();
	connect(bksl, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(bksl);

	//ROW 3
	x = stx;
	y += bksl->height() + vKeySpace;

	caps = new VButton(this);
	caps->setKeyCode(66);
	caps->resize(63, caps->height());
	caps->move(x, y);
	x += caps->width() + hKeySpace;
	//: button
	caps->setText(tr("CapsL"));
	caps->setCheckable(true);
	caps->res();
	otherKeys.append(caps);
	connect(caps, SIGNAL(clicked()), this, SLOT(toggleCaps()));

	//QString k3= "asdfghjkl"; //;'";
	//QString k3s="ASDFGHJKL";
	unsigned int kc3[ROWLENGTH3] = {38, 39, 40, 41, 42, 43, 44, 45, 46}; //,{47,48};

	for (int a = 0; a < ROWLENGTH3; a++) {
		VButton *v = new VButton(this);
		v->setKeyCode(kc3[a]);
		//v->setText ( k3.mid ( a,1 ) );
		//v->setShiftText ( k3s.mid ( a,1 ) );
		v->move(x, y);
		x += v->width() + hKeySpace;
		buttonsList.append(v);
		v->res();
		connect(v, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	}
	VButton *smcl = new VButton(this);
	smcl->setKeyCode(47);
	smcl->move(x, y);
	x += smcl->width() + hKeySpace;
	//smcl->setText ( ";" );
	//smcl->setShiftText ( ":" );
	connect(smcl, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(smcl);
	smcl->res();

	VButton *sngq = new VButton(this);
	sngq->setKeyCode(48);
	sngq->move(x, y);
	x += sngq->width() + hKeySpace;
	//sngq->setText ( "'" );
	//sngq->setShiftText ( "\"" );
	connect(sngq, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(sngq);
	sngq->res();

	VButton *enter = new VButton(this);
	enter->setKeyCode(36);
	enter->resize(58, enter->height());
	enter->move(x, y);
	x += enter->width() + hKeySpace;
	//: button
	enter->setText(tr("Enter"));
	connect(enter, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	otherKeys.append(enter);
	enter->res();

	//ROW 4
	x = stx;
	y += enter->height() + vKeySpace;

	leftShift = new VButton(this);
	leftShift->setKeyCode(50);
	leftShift->resize(80, leftShift->height());
	leftShift->move(x, y);
	x += leftShift->width() + hKeySpace;
	//: button
	leftShift->setText(tr("Shift"));
	leftShift->setCheckable(true);
	connect(leftShift, SIGNAL(toggled(bool)), this, SLOT(resetKeysCap()));
	modKeys.append(leftShift);
	leftShift->res();
	globalShift = leftShift;

	//QString k4="zxcvbnm"; //,./";
	//QString k4s="ZXCVBNM";
	unsigned int kc4[ROWLENGTH4] = {52, 53, 54, 55, 56, 57, 58};//59,60,61};

	for (int a = 0; a < ROWLENGTH4; a++) {
		VButton *v = new VButton(this);
		v->setKeyCode(kc4[a]);
		//v->setText ( k4.mid ( a,1 ) );
		//v->setShiftText ( k4s.mid ( a,1 ) );
		v->move(x, y);
		x += v->width() + hKeySpace;
		buttonsList.append(v);
		v->res();
		connect(v, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	}

	VButton *sm = new VButton(this);
	sm->setKeyCode(59);
	sm->move(x, y);
	x += sm->width() + hKeySpace;
	//sm->setText ( "," );
	//sm->setShiftText ( "<" );
	connect(sm, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(sm);
	sm->res();

	VButton *gr = new VButton(this);
	gr->setKeyCode(60);
	gr->move(x, y);
	x += gr->width() + hKeySpace;
	//gr->setText ( "." );
	//gr->setShiftText ( ">" );
	connect(gr, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(gr);
	gr->res();

	VButton *sl = new VButton(this);
	sl->setKeyCode(61);
	sl->move(x, y);
	x += sl->width() + hKeySpace;
	//sl->setText ( "/" );
	//sl->setShiftText ( "?" );
	connect(sl, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	buttonsList.append(sl);
	sl->res();

	rightShift = new VButton(this);
	rightShift->setKeyCode(62);
	rightShift->resize(77, rightShift->height());
	rightShift->move(x, y);
	x += rightShift->width() + hKeySpace;
	//: button
	rightShift->setText(fixedRightShift?tr("FixShift"):tr("Shift"));
	rightShift->setCheckable(true);
	connect(rightShift, SIGNAL(toggled(bool)), this, SLOT(resetKeysCap()));
	//connect(leftShift, SIGNAL(toggled(bool)), rightShift, SLOT(setChecked(bool)));
	//connect(rightShift, SIGNAL(toggled(bool)), leftShift, SLOT(setChecked(bool)));
	if (fixedRightShift) connect(rightShift, SIGNAL(clicked()), this, SLOT(rightShiftClicked())); //alex
	modKeys.append(rightShift);
	rightShift->res();
    ::rightShift = rightShift;

	//ROW 5
	x = stx;
	y += rightShift->height() + vKeySpace;

	leftCtrl = new VButton(this);
	leftCtrl->resize(45, leftCtrl->height());
	leftCtrl->move(x, y);
	x += leftCtrl->width() + hKeySpace;
	//: button
	leftCtrl->setText(tr("Ctrl"));
	leftCtrl->setKeyCode(37);
	leftCtrl->setCheckable(true);
	modKeys.append(leftCtrl);
	leftCtrl->res();


	winButton = new VButton(this);
	winButton->resize(45, winButton->height());
	winButton->move(x, y);
	x += winButton->width() + hKeySpace;
	//: button Windows(Super) key
	winButton->setText(tr("Win"));
    winButton->setKeyCode(133);    //winButton->setKeyCode(115);
#ifdef WIN_KEY_NO_MOD
    connect(winButton, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int))); //alex2
	otherKeys.append(winButton); //alex2: dont need to be modKeys
#else
    winButton->setCheckable(true);
    modKeys.append(winButton);
#endif
	winButton->res();


	leftAlt = new VButton(this);
	leftAlt->resize(45, leftAlt->height());
	leftAlt->move(x, y);
	x += leftAlt->width() + hKeySpace;
	//: button
	leftAlt->setText(tr("Alt"));
	leftAlt->setKeyCode(64);
	leftAlt->setCheckable(true);
	modKeys.append(leftAlt);
	leftAlt->res();


	VButton *space = new VButton(this);
	space->setKeyCode(65);
	space->resize(5*(space->width() + hKeySpace) /*+ 28*/, space->height());
	space->move(x, y);
	x += space->width() + hKeySpace;
	connect(space, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	space->res();
	otherKeys.append(space);

	rightAlt = new VButton(this);
	rightAlt->resize(45, rightAlt->height());
	rightAlt->move(x, y);
	x += rightAlt->width() + hKeySpace;
	//: button
	rightAlt->setText(tr("AltG"));
	//rightAlt->setKeyCode(113);
	rightAlt->setKeyCode(108);
	rightAlt->setCheckable(true);
	modKeys.append(rightAlt);
	rightAlt->res();


	menuButton = new VButton(this);
	menuButton->resize(45, menuButton->height());
	menuButton->move(x, y);
	x += menuButton->width() + hKeySpace;
	//: button Menu(Context Menu) key
	menuButton->setText(tr("Menu"));
	//menuButton->setKeyCode(117);
	menuButton->setKeyCode(135);
	menuButton->setCheckable(false);
	connect(menuButton, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	otherKeys.append(menuButton);
	menuButton->res();


	rightCtrl = new VButton(this);
	rightCtrl->resize(43, rightCtrl->height());
	rightCtrl->move(x, y);
	x += rightCtrl->width() + hKeySpace;
	//: button
	rightCtrl->setText(tr("Ctrl"));
	rightCtrl->setKeyCode(105);
	rightCtrl->setCheckable(true);
	modKeys.append(rightCtrl);
	rightCtrl->res();

	if (!funcRow) 
	{
		esc = new VButton(this);
		esc->setKeyCode(9);
		esc->resize(23 + 28, esc->height());
		esc->move(x, y);
		x += esc->width()+hKeySpace;
		//: button
		esc->setText(tr("Esc"));
		esc->res();
		otherKeys.append(esc);
		connect(esc, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	}
	else x += 23+28+hKeySpace;
	
	y = sty /*+ (funcRow?(rightCtrl->height() + vKeySpace):0)*/;

	quit = nullptr;
	//if (!isAlone())
	{
		quit = new ControlButton(this);
		quit->resize(27, quit->height() 
#ifdef WITH_SONNET
                     * (useSonnet?1:2) 
#else                
                     * 2
#endif
                     /*+ vKeySpace*/);
		quit->move(x, y);
		y += quit->height() + vKeySpace;
		quit->res();
		otherKeys.append(quit);

		if (QSystemTrayIcon::isSystemTrayAvailable())
		{
			///quit->setToolTip(tr("Minimize"));
			quit->setIcon(QIcon::fromTheme("go-down"));
			connect(quit, SIGNAL(clicked()), this, SLOT(hideByUserSlot()));
		}
		else
		{
			///quit->setToolTip(tr("Quit"));
			quit->setIcon(QIcon::fromTheme(/*"application-exit"*/"go-down"));
            //keep application run/closed or quit completely?
			connect(quit, SIGNAL(clicked()), this/*qApp*/, SLOT(close())/*SLOT(quit())*/);
		}
		quit->setIconSize(QSize(32, 32));

#ifdef WITH_SONNET
        if (useSonnet)
        {
            suggestMenu = new ControlButton(this);
            suggestMenu->resize(27, suggestMenu->height());
            suggestMenu->move(x, y);
            y += suggestMenu->height() + vKeySpace;
            suggestMenu->res();
            otherKeys.append(suggestMenu);

            suggestMenu->setIcon(QIcon::fromTheme(/*"tools-check-spelling"*/"applications-office"));
            suggestMenu->setIconSize(QSize(32, 32));

            //suggestMenu->setIcon(SmallIcon("configure"));
            ///suggestMenu->setToolTip(tr("Speller menu"));

            connect(suggestMenu,SIGNAL(clicked(bool)),this,SLOT(showSuggestTable(bool)));
        }
#endif

        if (settings.value("buttonMenu", true).toBool())
        {
            buttonMenu = new ControlButton(this);
            buttonMenu->resize(27, buttonMenu->height());
            buttonMenu->move(x, y);
            y += buttonMenu->height() + vKeySpace;
            buttonMenu->res();
            otherKeys.append(buttonMenu);

            buttonMenu->setIcon(QIcon::fromTheme(/*"preferences-other"*//*"preferences-system"*/"preferences-desktop-keyboard"));
            buttonMenu->setIconSize(QSize(32, 32));

            //buttonMenu->setIcon(SmallIcon("configure"));
            ///buttonMenu->setToolTip(tr("Configure"));
        }
	}
	//else y += 2*(rightCtrl->height() + vKeySpace);

	extent = new ControlButton(this);
	extent->resize(27, extent->height() * (funcRow?2:1) + (funcRow?vKeySpace:0));
	extent->move(x, y);
	y += extent->height() + vKeySpace;
	extent->setIcon(QIcon::fromTheme("view-right-new"));
	QSize smallIconSize(32, 32);
	extent->setIconSize(smallIconSize);
	//: tooltip Button for toggling numpad
	///extent->setToolTip(tr("Show Numeric Keypad"));
	extent->res();
	otherKeys.append(extent);
	connect(extent, SIGNAL(clicked()) , this, SLOT(toggleNumericPad()));

	layoutSwitcher = new ControlButton(this);
	layoutSwitcher->resize(27, layoutSwitcher->height());
	layoutSwitcher->move(x, y /*+ (layoutSwitcher->height() + vKeySpace) / 2*/);
	y += layoutSwitcher->height() + vKeySpace;
	layoutSwitcher->res();
	layoutSwitcher->setIconSize(smallIconSize);
	otherKeys.append(layoutSwitcher);
    connect(layoutSwitcher, SIGNAL(clicked()), this, SLOT(changeLayout()));
	XkbStateRec xkbStateRec;
	XkbGetState(QX11Info::display(), XkbUseCoreKbd, &xkbStateRec);
	mappingNotify(xkbStateRec.group);

    modTimer = NULL;
    int modTimerInterval = settings.value("ModTimerInt","0").toInt(); //0 - default, <0 - dont use
    if (modTimerInterval==0) modTimerInterval = 1000; //was 500
    if (modTimerInterval>0)
    {
        modTimer = new QTimer(this);
        connect(modTimer, SIGNAL(timeout()), this, SLOT(queryModState()));
        modTimer->setInterval(modTimerInterval);
        modTimer->start();
    }

	int padx = 560;

	x = padx;
	y = sty + (funcRow?(30 + vKeySpace):0);

    bool vertical = screenGeometry.width()<screenGeometry.height()?true:false;

	numLock = new VButton(this);
	numLock->setKeyCode(77);
	numLock->resize(32, numLock->height());
	numLock->move(x, y);
	x += numLock->width() + hKeySpace;
	numLock->res();
	//: button Numpad numeric lock key
	numLock->setText(vertical?tr("Num\nLock"):tr("NumL"));
	numLock->setCheckable(true);
	otherKeys.append(numLock);
	connect(numLock, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
	connect(numLock, SIGNAL(clicked()), this, SLOT(toggleNumlock()));

	div = new VButton(this);
	div->resize(32, div->height());
	div->move(x, y);
	x += div->width() + hKeySpace;
	div->res();
	div->setText("/");
	div->setKeyCode(106);
	otherKeys.append(div);
	connect(div, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	mul = new VButton(this);
	mul->resize(32, mul->height());
	mul->move(x, y);
	x += mul->width() + hKeySpace;
	mul->res();
	mul->setText("*");
	mul->setKeyCode(63);
	otherKeys.append(mul);
	connect(mul, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	min = new VButton(this);
	min->resize(32, min->height());
	min->move(x, y);
	min->setText("-");
	min->setKeyCode(82);
	otherKeys.append(min);
	min->res();
	connect(min, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	int y_plus = y + numLock->height() + vKeySpace;

	QString txt[9] = { vertical?tr("Ho\nme"):tr("Home"), QString::fromUtf8("▲"), vertical?tr("Pg\nUp"):tr("PgUp"),
			   QString::fromUtf8("◄"), " ", QString::fromUtf8("►"), tr("End"),
			   QString::fromUtf8("▼"), vertical?tr("Pg\nDn"):tr("PgDn")
			 };
	//FIXME: should this be translatable?
	QString nump[9] = {"7", "8", "9", "4", "5", "6", "1", "2", "3"};
	int val = 0;
	int nval[9] = {79, 80, 81, 83, 84, 85, 87, 88, 89};
	for (int a = 2;a < 5; a++) {
		x = padx;
		y += numLock->height() + vKeySpace;

		for (int b = 0;b < 3;b++) {
			NumpadVButton *v = new NumpadVButton(this);
			v->resize(32, v->height());
			v->move(x, y);
			x += v->width() + hKeySpace;
			v->res();
			v->setKeyCode(nval[val]);
			v->setText(txt[val], nump[val]);
			numLockKeys.append(v);
			connect(v, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
			val++;
		}
	}

	plu = new VButton(this);
	plu->resize(32, plu->height()*2 + vKeySpace);
	plu->move(x, y_plus);
	plu->res();
	plu->setText("+");
	plu->setKeyCode(86);
	otherKeys.append(plu);
	connect(plu, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	ent = new VButton(this);
	ent->resize(32, ent->height()*2 + vKeySpace);
	ent->move(x, y_plus + plu->height() + vKeySpace);
	ent->res();
	//: button Numpad enter key
	ent->setText(tr("Ent"));
	ent->setKeyCode(104);
	otherKeys.append(ent);
	connect(ent, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	x = padx;
	y += numLock->height() + vKeySpace;

	ins = new NumpadVButton(this);
	ins->resize(numLock->width()*2 + hKeySpace, ins->height());
	ins->move(x, y);
	x += ins->width() + hKeySpace;
	ins->res();
	//: button Numpad insert key
	ins->setText(tr("Ins"), "0");
	ins->setKeyCode(90);
	numLockKeys.append(ins);
	connect(ins, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));

	del = new NumpadVButton(this);
	del->resize(32, del->height());
	del->move(x, y);
	del->res();
	//: button Numpad delete key
	del->setText(tr("Del"), ".");
	del->setKeyCode(91);
	numLockKeys.append(del);
	connect(del, SIGNAL(keyClick(unsigned int)), this, SLOT(keyPress(unsigned int)));
}

void MainWidget::rightShiftClicked()
{
	if (!rightShift || !fixedRightShift) return;

	fixedRightShift=false;

	if (rightShift->isChecked() && caps && caps->isChecked())
	{
		caps->setChecked(false);
		toggleCaps();
		//sendKey(caps->getKeyCode());
		//resetKeysCap();
	}

	if (numLock) 
	{
		;//FIXME: if keep Shift pressed then numbers will be printed and vice versa
	}

	fixedRightShift=true;
}

KeySym MainWidget::keyCodeToKeySym(KeyCode keyCode, int group, uint modsMask)
{
	XkbDescRec *xkb = getXkbDesc();
	uint mask = modsMask;

    if (!xkb) return 0;

	int type =  xkb->map->key_sym_map[keyCode].kt_index[group];
	int level = keyLevelForType(QX11Info::display(), xkb, type, &mask);

	KeySym keysym = XkbKeycodeToKeysym(QX11Info::display(), keyCode, group, level);

	//HACK: XKeycodeToKeysym report weird values for "^" and "¨"
	//We should consider using another way to get the events if it happens for other keys
	if (keysym == 0xfe52) {
		keysym = 0x005e;
	} else if (keysym == 0xfe57) {
		keysym = 0x00a8;
	}

	return keysym;
}

XkbDescPtr MainWidget::getXkbDesc()
{
	if (m_XkbDescPtr == NULL)
	{
		/*int ir, reason_return;
		char *display_name = getenv ("DISPLAY");
		m_XkbDisplay = XkbOpenDisplay (display_name,
					       &gok_xkb_base_event_type,
					       &ir, NULL, NULL,
					       &reason_return);
		if (m_XkbDisplay == NULL)
			qWarning() << QString("Xkb extension could not be initialized! (error code %1)").arg(reason_return);
		else */
			m_XkbDescPtr = XkbGetMap (QX11Info::display(),
						  XkbAllComponentsMask,
						  XkbUseCoreKbd);
		if (m_XkbDescPtr == NULL)
			qWarning() << "keyboard description not available!";
		else {
			//int status = XkbGetGeometry (/*m_XkbDisplay*/QX11Info::display(), m_XkbDescPtr);

			/*if (status != Success)
			{
				qWarning() << "Keyboard Geometry cannot be read from your X Server.";
			}*/
			XkbGetNames (QX11Info::display()/*m_XkbDisplay*/, XkbAllNamesMask, m_XkbDescPtr);
		}
	}
	return m_XkbDescPtr;
}

int MainWidget::keyLevelForType(Display * /*display*/, XkbDescRec * kbd, int type, uint* modmask)
{
	int i, level = 0;
	bool level_set = false;
	XkbKeyTypeRec *key_type = &kbd->map->types[type];

	/* compare against each map entry in the XkbKeyType */
	for (i = 0; i < key_type->map_count && !level_set; ++i)
	{
		unsigned int vmods_equiv = 0, level_mods = 0;
		unsigned int mods_mask;
		unsigned int mods = *modmask;
		XkbVirtualModsToReal (kbd, key_type->map[i].mods.vmods,
				      &vmods_equiv);
		XkbVirtualModsToReal (kbd, key_type->mods.vmods,
				      &mods_mask);
		mods_mask = key_type->mods.mask;
		level_mods = key_type->map[i].mods.real_mods | vmods_equiv;
		if ((key_type->map[i].active) &&
		    ((mods & mods_mask) == level_mods))
		{
			level = key_type->map[i].level;
			level_set = /*TRUE*/1;
			/**
			 *  Special hack for alpha types: see section 15.2.1 of XKBlib spec.
			 *  We'd restrict this to the canonical types, but XFree's maps tend to use
			 *  non-canonical types for the primary alphanumeric keycodes, in order
			 *  to support more shift levels (AltGr, Shift-AltGr, etc.)
			 *  If we reported the Xkb 'level' here, the keycap strings would not get
			 *  'shifted' in our compose keyboard when CapsLock is active, since the
			 *  capitalization in that case is deferred to the Xlib lookup stage.
			 **/
			if (key_type->preserve && (key_type->preserve[i].mask == LockMask))
			{
			    /* Lock is preserved for Xlib, but we want to report 'equivalent' level */
			    if ((mods & LockMask) && !(mods & ShiftMask))
			    {
				int j;
				/* find the map matching the 'Shift' equivalent */
				uint equiv_mods = (mods & ~LockMask) | ShiftMask;
				for (j = 0; j < key_type->map_count; ++j)
				{
				    if (key_type->map[j].active &&
					(key_type->map[j].mods.real_mods == equiv_mods))
				    {
					level = key_type->map[j].level;
					level_set = /*TRUE*/1;
					break;
				    }
				}
			    }
			}

			/* preserve what needs preserving */
			if (key_type->preserve)
				*modmask = *modmask & key_type->preserve[i].mask;
			else
				*modmask = 0;
			break;
		}
	}
	if (i == key_type->map_count) /* no explicit match */
		*modmask = 0;

	return level;
}

void MainWidget::setStrutGeometry() //alex: it has effect only if no bypass fly-wm
{
	if (!strut) return;

	//desktop->blockSignals(true);

	qDebug() << Q_FUNC_INFO;
/*
#ifdef NO_KWINDOWSYSTEM
	static Atom XA_NET_WM_STRUT = None;
	if (XA_NET_WM_STRUT == None)
		XA_NET_WM_STRUT = XInternAtom(display, "_NET_WM_STRUT", False);
#endif
*/
	//it has to be placed along screen edge
	QRect newg = geometry();

    //QRect scrg = desktop->screenGeometry(this);
    QDesktopWidget *dw = QApplication::desktop();
    QRect scrg = dw->screenGeometry(dw->primaryScreen());

	//fprintf(stderr,"fly-vkbd: new screen geom = %ld %ld %ld %ld strut=%d\n",scrg.x(),scrg.y(),scrg.width(),scrg.height(),strut);

	if (strut==StrutCurrent) 
	{
		if (geometry().x() > scrg.x()-10 && geometry().x() < scrg.x()+10) newg.setX(scrg.x());
		if (geometry().y() > scrg.y()-10 && geometry().y() < scrg.y()+10) newg.setY(scrg.y());

		if (newg.x()+newg.width() > scrg.x()+scrg.width()-10 && newg.x()+newg.width() < scrg.x()+scrg.width()+10)
			newg.setWidth(scrg.x()+scrg.width() - newg.x());

		if (newg.y()+newg.height() > scrg.y()+scrg.height()-10 && newg.y()+newg.height() < scrg.y()+scrg.height()+10)
			newg.setHeight(scrg.y()+scrg.height() - newg.y());
        //fprintf(stderr,"fly-vkbd: curr geom is %ld %ld %ld %ld whdl=%x\n",newg.x(),newg.y(),newg.width(),newg.height(),windowHandle());
    }
	else if (strut==StrutBottom) 
	{
        int strutHeight;
        int strutY;

        if (!geom.isNull() && geom.height() &&  geom.y()>=scrg.y() && geom.y()<scrg.y()+scrg.height())
        {
            strutHeight = geom.height();
            strutY =geom.y();
        }
        else
        {
          strutHeight = (strutBottomHeight>1.0? strutBottomHeight:((double)scrg.height()*strutBottomHeight));
          if (strutHeight >= scrg.height()) strutHeight = scrg.height()/3; //some fallback

          strutY = scrg.y() + scrg.height() - strutHeight;
        }

		newg.setX(scrg.x());
		newg.setY(strutY);
		newg.setWidth(scrg.width()); 
		newg.setHeight(strutHeight);
        //fprintf(stderr,"fly-vkbd: bottom geom is %ld %ld %ld %ld whdl=%x\n",newg.x(),newg.y(),newg.width(),newg.height(),windowHandle());
	}

	setGeometry(newg);
	if (windowHandle()) windowHandle()->setGeometry(newg);

	qDebug() << "newg = " << newg;

	if (strut!=StrutNone)
	{
	//inform WM about strut
	long val[4]={0,0,0,0};
	
	if (strut==StrutBottom)
	{
        if (!geom.isNull() && geom.height())
        {
            val[3]=scrg.y()+scrg.height() - geom.y();
            //fprintf(stderr,"scr height=%d gem=(%d %d %d %d) gmetry=(%d %d %d %d) strut=%d\n",scrg.height(),
              //      geom.x(),geom.y(),geom.width(),geom.height(),
                //    geometry().x(),geometry().y(),geometry().width(),geometry().height(),
                  //  val[3]);
        }
        else
        {
            val[3]=geometry().height();
        }
	}
	else
	{
	    if (geometry().y()==0) val[2] = geometry().height();//top
	    else val[3]=geometry().height(); //bottom
	}
//#ifndef NO_KWINDOWSYSTEM
    const auto left = val[0];
    const auto right = val[1];
    const auto top = val[2];
    const auto bottom = val[3];
    const auto w = scrg.width();
    const auto h = scrg.height();
    KWindowSystem::setExtendedStrut(winId(), left, 0, left != 0 ? w : 0, right, 0, right != 0 ? w : 0,
                                    top, 0, top != 0 ? h : 0, bottom, 0, bottom != 0 ? h : 0);
/*#else
	static Atom XA_NET_WM_STRUT = None;
	if (XA_NET_WM_STRUT == None)
		XA_NET_WM_STRUT = XInternAtom(display, "_NET_WM_STRUT", False);
	XChangeProperty(display, winId(),
			XA_NET_WM_STRUT, XA_CARDINAL, 32,  PropModeReplace, (unsigned char *)val, 4);
#endif*/
	//fprintf(stderr,"fly-vkbd: %lx strut is %ld %ld %ld %ld\n",winId(),val[0],val[1],val[2],val[3]);
	}

	//QApplication::syncX(); //XSync(display,false); - really need?
	//desktop->blockSignals(false);
}

// TODO: find a way to hide the widget before asking for quit
// FIXME: clicking on the tray icon to hide/show the widget only work for showinButtong it
KbdTray::KbdTray(MainWidget* parent) : QSystemTrayIcon(/*parent*/)
{
	p=parent;
	setObjectName("Kvkbd Tray");

    QIcon icon = QIcon::fromTheme("input-keyboard");

	XkbStateRec xkbStateRec;
	XkbGetState(QX11Info::display(), XkbUseCoreKbd, &xkbStateRec);
	QString layout = parent->getLayouts().value(xkbStateRec.group);

    QString layoutIconName = parent->iconNameForLayout(layout);
    if (!QIcon::hasThemeIcon(layoutIconName)) layoutIconName = QString("flag-") + layout;

    QIcon iconTray = FlyIcon::withOverlay(icon, Qt::BottomRightCorner, 
                                          QIcon::fromTheme(layoutIconName), emblemSize);
    setIcon(iconTray);

    setToolTip(tr("Virtual keyboard"));

	/*QAction *titleAction = contextMenuTitle();
	titleAction->setText("Kvkbd");
	setContextMenuTitle(titleAction);*/

	connect(this, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(triggerParentWidgetVisibility(QSystemTrayIcon::ActivationReason)));
}

KbdTray::~KbdTray()
{
}

void KbdTray::triggerParentWidgetVisibility(QSystemTrayIcon::ActivationReason reason)
{
	if (reason == QSystemTrayIcon::Trigger)
	{
		//QWidget *p = qobject_cast<QWidget*>(parent());
		if (p)
		{
            p->setVisibilityForced(!p->isVisible());
			p->setVisible(!p->isVisible());
            p->raise();
		}
	}
}
/*
extern "C" {
static int IgnoreXError(Display *, XErrorEvent *) {
	return 0;
}
}
*/
QStringList MainWidget::getLayouts() const
{
	QStringList groups;

	Display* dpy = QX11Info::display();

	static Atom rules_atom = None;

    if (rules_atom == None) rules_atom = XInternAtom(dpy, _XKB_RF_NAMES_PROP_ATOM, False);
    if (rules_atom == None) /* property cannot exist */
	{
		qCritical() << "Failed to fetch layouts from server:" << "could not find the atom" << _XKB_RF_NAMES_PROP_ATOM;
		return groups;
	}

	Atom real_prop_type;
	int fmt;
	unsigned long nitems;
	unsigned long extra_bytes;
	char *prop_data = NULL;

	Status ret = XGetWindowProperty(dpy,
					QX11Info::appRootWindow(),
					rules_atom, 0L, _XKB_RF_NAMES_PROP_MAXLEN,
					False, XA_STRING, &real_prop_type, &fmt,
					&nitems, &extra_bytes,
					( unsigned char ** ) ( void * ) & prop_data );

	/* property not found! */
	if ( ret != Success )
	{
		qCritical() << "Failed to fetch layouts from server:" << "Could not get the property";
		return groups;
	}

	/* has to be array of strings */
	if ( ( extra_bytes > 0 ) || ( real_prop_type != XA_STRING )
			|| ( fmt != 8 ) )
	{
		if (prop_data)
			XFree(prop_data);
		qCritical() << "Failed to fetch layouts from server:" << "Wrong property format";
		return groups;
	}

	// skip over rules
	char* out = prop_data;
	out += strlen(out) + 1;

	// skip over model
	if (out - prop_data < nitems) {
		out += strlen(out) + 1;
	}

	// get layout
	if ((unsigned long)(out - prop_data) < nitems && *out) {
		groups = QString::fromLocal8Bit(out).split(',');
	}

	XFree ( prop_data );
	return groups;
}

int MainWidget::changeLayoutTo(const QString & lname, const QString & alias, int *old)
{
	QStringList layouts = getLayouts();

	if (layouts.size() > 1)
	{
		XkbStateRec xkbStateRec;
		XkbGetState(QX11Info::display(), XkbUseCoreKbd, &xkbStateRec);
		int i = xkbStateRec.group;

        int j;
        for (j = 0; j < layouts.size(); j++) 
        {
            if (layouts.at(j).contains(lname,Qt::CaseInsensitive) || 
                layouts.at(j).contains(alias,Qt::CaseInsensitive)) break;
        }
		if (j < layouts.size() && i!=j) 
        {
            XkbLockGroup(QX11Info::display(), XkbUseCoreKbd, j);
            if (old) *old=i;
            return j;
        }
	}
    return -1;
}

void MainWidget::changeLayout()
{
	QStringList layouts = getLayouts();

	if (layouts.size() > 1)
	{
		XkbStateRec xkbStateRec;
		XkbGetState(QX11Info::display(), XkbUseCoreKbd, &xkbStateRec);
		int i = xkbStateRec.group + 1;

		if (i >= layouts.size()) i = 0;

		XkbLockGroup(QX11Info::display(), XkbUseCoreKbd, i);

        mappingNotify(i); //call now not to wait mapping notify
	}
}

QString MainWidget::iconNameForLayout(const QString &layout) const
{
    if (layout=="ru") 
    {
      if (QIcon::hasThemeIcon("keyboard-layout-ru-red")) return QString("keyboard-layout-ru-red");
      return QString("keyboard-layout-ru");
    }
    else if (layout=="en" || layout=="us") return QString("keyboard-layout-en");

    return QString("keyboard-layout-"+layout);
}

QIcon MainWidget::iconForLayout(const QString &layout) const
{
	QIcon ico;
    QString iconName = iconNameForLayout(layout);

    ico = QIcon::fromTheme(iconName);
	if (ico.isNull()) ico = QIcon::fromTheme("flag-" + layout);

	return ico;
}

QMenu* MainWidget::createMenu()
{
	QSettings s;

	QMenu* menu = new QMenu("fly-vkbd", this);

	chooseFontAction = new QAction(QIcon::fromTheme("preferences-desktop-font"), tr("Choose Font..."), this);
	connect(chooseFontAction, SIGNAL(triggered(bool)), this, SLOT(chooseFont()));

	//chooseBackgroundColorAction = new QAction(tr("Choose background color..."), this);
	//connect(chooseBackgroundColorAction, SIGNAL(triggered( bool )), this, SLOT(chooseBackgroundColor()));

	QAction *autoResizeAction = new QAction(tr("Auto Resize Font"), this);
	autoResizeAction->setCheckable(true);

	autoRepeated = s.value("autoRepeated", true).toBool();
	QAction *autoRepeatedAction = new QAction(tr("Autorepeat"), this);
	autoRepeatedAction->setCheckable(true);
	autoRepeatedAction->setChecked(autoRepeated);

    QAction *funcRowAction = new QAction(tr("Show F1-F12"), this);
    funcRowAction->setCheckable(true);
    funcRowAction->setChecked(funcRow);

#ifdef WITH_SONNET
    useSonnet = s.value("useSonnet", false).toBool();
	QAction *suggestAction = new QAction(tr("Suggestions"), this);
	suggestAction->setCheckable(true);
	suggestAction->setChecked(useSonnet);
	connect(suggestAction, SIGNAL(toggled(bool)), this, SLOT(useSuggestSlot(bool)));
#endif

#ifdef USE_DOCK
	QAction *showDockAction = new QAction(tr("Dock Keyboard"), this);
	showDockAction->setCheckable(true);
	showDockAction->setChecked(s.value("showdock", false).toBool());
#endif

	QAction *lockWidgetAction = new QAction(tr("Lock on Screen"), this);
	lockWidgetAction->setCheckable(true);
	lockWidgetAction->setChecked(s.value("locked", false).toBool());

#ifdef WITH_SOLID
    QAction* autoShowAction = new QAction(tr("Autoshow if no real keyboard"), this);
#else
    QAction* autoShowAction = new QAction(tr("Autoshow"), this);
#endif
    autoShowAction->setCheckable(true);

    flySession = QString::fromLocal8Bit(flySessionName());
    //fprintf(stderr,"sess=%s\n",flySession.toLocal8Bit().data());

    autoShow = s.value(flySession+"/autoShow",whatFlySession()==FLY_MOBILE?false:true).toBool();

    autoShowAction->setChecked(autoShow);


	QAction* quitAction = new QAction(QIcon::fromTheme("application-exit"), tr("&Quit"), this);

	connect(this, SIGNAL(autoResizeToggled(bool)), autoResizeAction, SLOT(setChecked(bool)));
	connect(autoResizeAction, SIGNAL(triggered(bool)), this, SLOT(autoResizeFont(bool)));

	connect(autoRepeatedAction, SIGNAL(toggled(bool)), this, SLOT(autoRepeatSlot(bool)));

	connect(funcRowAction, SIGNAL(toggled(bool)), this, SLOT(funcRowSlot(bool)));

	//connect(dock, SIGNAL(dockShown(bool)), showDockAction, SLOT(setChecked(bool)));
#ifdef USE_DOCK
	connect(showDockAction, SIGNAL(toggled(bool)), this, SLOT(showDock(bool)));
#endif
	connect(this, SIGNAL(widgetLocked(bool)), lockWidgetAction, SLOT(setChecked(bool)));
	connect(lockWidgetAction, SIGNAL(toggled(bool)), this, SLOT(lockWidget(bool)));

    connect(autoShowAction, SIGNAL(toggled(bool)), this, SLOT(autoShowSlot(bool)));

	connect(quitAction, SIGNAL(triggered( bool )), qApp, SLOT(quit()));

	FlyHelpMenu *helpMenu = new FlyHelpMenu(this);

	if (chooseFontAction)               menu->addAction(chooseFontAction);
	//if (chooseBackgroundColorAction)    menu->addAction(chooseBackgroundColorAction);
	if (autoResizeAction)               menu->addAction(autoResizeAction);
	menu->addSeparator();
	if (autoRepeatedAction)             menu->addAction(autoRepeatedAction);
	if (funcRowAction)                  menu->addAction(funcRowAction);
#ifdef USE_DOCK
	if (showDockAction)                 menu->addAction(showDockAction);
#endif
	if (lockWidgetAction)               menu->addAction(lockWidgetAction);
	if (autoShowAction)                 menu->addAction(autoShowAction);
#ifdef WITH_SONNET
	if (suggestAction)                  menu->addAction(suggestAction);
#endif

	helpAction = menu->addMenu(helpMenu);
	QAction *a = helpMenu->actions()[0]; //helpMenu->actionAt(QPoint(50,30)); - failed probale because no window yet
	if (a) a->setShortcut(QKeySequence());

	menu->addSeparator();
	if (quitAction) menu->addAction(quitAction);

	return menu;
}
void MainWidget::setCssStyle(const QString& s)
{
	cssStyle = (s=="restore"? cssStyleOrig : s);

	QFont f = font();
	updateFont(f);

	if (strut) //alex: to fit font
	{
	    QResizeEvent re(size(),size());
	    resizeEvent(&re);
	    repaint();
	}
}

#ifdef QAccessibilityClient_FOUND

void MainWidget::focusChanged(const QAccessibleClient::AccessibleObject &object)
{
    static int debug_print=-1;
    if (debug_print==-1) debug_print=(getenv("FLY_VKBD_PRINT_FOCUS_ITEMS")?1:0);

    if (showHideTimer->isActive()) showHideTimer->stop();

#ifdef WITH_SOLID
    if (autoShow && isKbdAttached) return;
#else
    if (!autoShow) return;
#endif

    QAccessibleClient::AccessibleObject checkObject;
    if (object.roleName()=="combo box" && !object.isEditable() && object.childCount()>1) 
         checkObject = object.child(1);
    else checkObject = object;

    //QPoint                        pFocus    = checkObject.focusPoint();
	QString 						roleName  = checkObject.roleName();
	QString 						lroleName = checkObject.localizedRoleName();
	//AccessibleObject::Role		role      = checkObject.role();
	AccessibleObject::Interfaces	ifaces    = checkObject.supportedInterfaces();

#ifdef WITH_SONNET
    QAccessibleClient::AccessibleObject parentObject = object.parent();
    if (parentObject.isValid())
    {
        QString parentName = parentObject.name();
        if (parentName==FLY_VKBD_LIST_WIDGET_NAME)
        {
          if (debug_print)
            fprintf(stderr,"fly-vkbd %s: objName=%s,desc=%s,id=%s,url=%s,parent=%s role='%s' lrole='%s', ifaces=%o(%x)  editabletxt=%o(%x) aboutToVisible=%d winsize=%dx%d text='some txt length=%lu'\n",
              aboutToVisible?"show":"hide", object.name().toLocal8Bit().data(), object.description().toLocal8Bit().data(),
              object.id().toLocal8Bit().data(), object.url().toString().toLocal8Bit().data(),
              object.parent().name().toLocal8Bit().data(),
              roleName.toLocal8Bit().data(),lroleName.toLocal8Bit().data(),
              (int)ifaces,(int)ifaces,AccessibleObject::EditableTextInterface,AccessibleObject::EditableTextInterface,
              aboutToVisible,width(),height(),strlen(object.text().toLocal8Bit().data()));
          return;
        }
    }
#endif
	if (focusRoles.contains(roleName,Qt::CaseInsensitive)
		   || (ifaces & AccessibleObject::EditableTextInterface)
		// || object.isEditable()
		// || (ifaces & AccessibleObject::TextInterface)
		// || (role & AccessibleObject::Text)
		// || (role & AccessibleObject::PasswordText)
		// || (role & AccessibleObject::ComboBox)
		// || (role & AccessibleObject::TableCell)
		)
    {
             if (roleName=="combo box" && !checkObject.isEditable())  aboutToVisible=false;
        else if (focusNoRoles.contains(roleName,Qt::CaseInsensitive)) aboutToVisible=false;
        else aboutToVisible=true;
    }
    else {
        if (isVisible() && roleName=="menu item") aboutToVisible=true;
#ifdef WITH_SONNET
        else if (isVisible() && !parentObject.isValid()) aboutToVisible=true;
#endif
        else aboutToVisible=false;
    }

    if (debug_print)
      fprintf(stderr,"fly-vkbd %s: objName=%s,desc=%s,id=%s,url=%s,parent=%s('%s') role='%s' lrole='%s', ifaces=%o(%x)  editabletxt=%o(%x) aboutToVisible=%d winsize=%dx%d text='some txt length=%lu'\n",
              aboutToVisible?"show":"hide", object.name().toLocal8Bit().data(), object.description().toLocal8Bit().data(),
              object.id().toLocal8Bit().data(), object.url().toString().toLocal8Bit().data(),

              parentObject.isValid()?"yes":"no",
              parentObject.isValid()?object.parent().name().toLocal8Bit().data():"noname",

              roleName.toLocal8Bit().data(),lroleName.toLocal8Bit().data(),
			  (int)ifaces,(int)ifaces,AccessibleObject::EditableTextInterface,AccessibleObject::EditableTextInterface,
              aboutToVisible,width(),height(),strlen(object.text().toLocal8Bit().data()));

	if (isVisible()!=aboutToVisible) showHideTimer->start(autoShowDelay);
}

void MainWidget::showHideTimeout()
{
    if (visibilityForced && !aboutToVisible) return; //dont hide if shown by user

	setVisible(aboutToVisible);
}
#endif

void MainWidget::setInLocker(bool lockerIsRun)
{
	if (buttonMenu) 
    {
        //buttonMenu->setEnabled(!lockerIsRun);
        if (lockerIsRun) buttonMenu->hide();
        else             buttonMenu->show();
    }
#ifdef WITH_SONNET
	if (suggestMenu) 
    {
        //suggestMenu->setEnabled(!lockerIsRun);
        if (lockerIsRun) suggestMenu->hide();
        else             suggestMenu->show();
    }
#endif
}
