/*
 * This file is part of the Kvkbd project.
 * Copyright (C) 2007-2008 Todor Gyumyushev <yodor@developer.bg>
 * Copyright (C) 2008 Guillaume Martres <smarter@ubuntu.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef KVKBDMAINWIDGET_H
#define KVKBDMAINWIDGET_H

#include <X11/Xlib.h>
#include <X11/extensions/XKBstr.h>
#include "fixx11h.h"

#include <QList>
#include <QSystemTrayIcon>
#include <QClipboard>

#ifdef QAccessibilityClient_FOUND
#include <qaccessibilityclient/registry.h>
#endif
#include "resizabledragwidget.h"

#ifdef WITH_SONNET
#include <sonnet/speller.h>
#include <sonnet/guesslanguage.h>
#include <QListWidget>
//class QTableWidget;
//class QListWidget;
//class QListWidgetItem;
#endif


class KAction;
class KAboutData;

class KbdDock;
class KbdTray;
class VButton;
class ControlButton;
class NumpadVButton;
class QDesktopWidget;
class QTimer;

#ifdef WITH_SONNET
class QToolButton;
class MyListWidget : public QListWidget
{
	Q_OBJECT
public:
    MyListWidget(QSize shrinkedSize, QSize expandedSize, QPoint pos, QWidget *parent = nullptr);
    ~MyListWidget();

    bool isShrinked() { return shrinked; }

private:
    QSize sSize, eSize;
    bool  shrinked;
    QToolButton *shrinkExpandBtn;
                   
public slots:
    void shrinkExpandSlot();
};
#endif

class MainWidget : public ResizableDragWidget
{
	Q_OBJECT
	Q_CLASSINFO("D-Bus Interface", "org.kde.kvkbd.Kvkbd")
	//Q_PROPERTY(bool alone READ isAlone)
	Q_PROPERTY(bool autoResize READ autoResized WRITE setAutoResized)

public:
	enum StrutType { StrutNone=0, StrutCurrent, StrutBottom };

	explicit MainWidget(StrutType strut, 
						const QColor &color, const QString& styleSheet = QString(),
						/*bool loginhelper,*/ 
                        QWidget *parent = nullptr, Qt::WindowFlags f = Qt::X11BypassWindowManagerHint);
	~MainWidget();

	void mappingNotify(/*XMappingEvent *e*/int group);
	// const KAboutData *aboutData;
	void saveState();
	//bool isAlone() const;
	bool autoResized() const;
	void setAutoResized(bool);

    void setVisibilityForced(bool v) { visibilityForced=v; }
    bool isVisibilityForced() { return visibilityForced; }

    QRect getScreenGeometry() { return screenGeometry; }
    void  setArgGeometry(const QString &geom);

//-------------------------------------------------
	QStringList getLayouts() const;
	QIcon       iconForLayout(const QString& layout) const;
    QString     iconNameForLayout(const QString &layout) const;
//-------------------------------------------------

    KbdTray *kbdTray() { return tray; }

	KbdDock *dock;

public slots:
	void autoResizeFont(bool);
	void chooseFont();
	//void chooseBackgroundColor();
	void lockWidget(bool);
	void showDock(bool);

	void workAreaResized(int);//alex
    void workAreaResized(const QRect & rect);//alex
	void screenCountChanged(int);
	void primaryScreenChanged();

	void autoRepeatSlot(bool);//alex
    void funcRowSlot(bool); //alex
    void setCssStyle(const QString&); //alex
    void setInLocker(bool);//alex
    void hideByUserSlot();
    void autoShowSlot(bool);
#ifdef WITH_SOLID
    void deviceChangedSlot(const QString&);
#endif
signals:
	void autoResizeToggled(bool);

private slots:
	void keyPress(unsigned int keycode);

	void toggleCaps();
	void toggleNumlock();

	void resetKeysCap();

	void queryModState();

	void toggleNumericPad();
	void restorePosition();

    int changeLayoutTo(const QString & lname, const QString & alias, int *old);//alex
	void changeLayout();

#ifdef QAccessibilityClient_FOUND
    void focusChanged(const QAccessibleClient::AccessibleObject &object); // Called from a dbus service
#endif

    void rightShiftClicked(); //alex

protected:
	void resizeEvent(QResizeEvent *e) override;
	void closeEvent(QCloseEvent *e) override;

private:
	void calculateSize();

	void finishInit();
	void setupKeyboard();
	void restoreSettings();

	void updateFont(QFont &font);

	bool resizable;

	bool keyState(int iKey);
	//void keyPress(int keycode); alex: unused, the copy of keyPress(unsigned int) above?
	void setupText(VButton &v, int group);

	void setStrutGeometry();

	void setBackgroundColor(const QColor& color);

	QString setFontInStyle(const QString& cssStyle, const QFont& fnt);

//-------------------------------------------------
	KeySym keyCodeToKeySym(KeyCode keyCode, int group, uint modsMask);
	XkbDescPtr getXkbDesc();
	int keyLevelForType(Display *display, XkbDescRec *kbd, int type, uint *modmask);

	XkbDescPtr m_XkbDescPtr;

    QMenu* createMenu();

	QTimer *modTimer;

	bool funcRow, autoShow;

	VButton *esc;

    VButton *bksp;

	VButton *caps;
	VButton *leftShift;
	VButton *rightShift;

	VButton *leftCtrl;
	VButton *rightCtrl;

	VButton *leftAlt;
	VButton *rightAlt;

	VButton *winButton;
	VButton *menuButton;

	VButton *numLock;
	VButton *div;
	VButton *mul;
	VButton *min;
	VButton *plu;
	VButton *ent;

	NumpadVButton *ins;
	NumpadVButton *del;

	QList<VButton *> buttonsList;
	QList<VButton *> modKeys;
	QList<VButton *> otherKeys;
	QList<NumpadVButton *> numLockKeys;

	void sendKey(unsigned int keycode);

	Display *display;
	KbdTray *tray;

	bool isExtentVisible;

	ControlButton *quit;
	ControlButton *buttonMenu;

    ControlButton *extent;
	ControlButton *layoutSwitcher;

	double sdxs;
	double sdxb;

	bool autoResize;
	//bool alone;
        double dpiMul;

	StrutType strut;
	double strutBottomHeight; //either in pixels (>1.0) or in percents of screen height (if <1.0)
    QRect geom;               //height from command line

	QDesktopWidget * desktop;
	int screenNum;
	QRect screenGeometry;

    QString cssStyle, cssStyleOrig;
    QColor  bgColor;

    QMenu *menu;
	QAction *chooseFontAction, /**chooseBackgroundColorAction,*/ *helpAction;

    QString flySession;

#ifdef WITH_SOLID
    bool isKbdAttached;
    bool findKbdAttached();
#endif

    bool visibilityForced;

#ifdef QAccessibilityClient_FOUND
    QAccessibleClient::Registry *accessibilityRegistry; // Global Accessibility Registry
	bool    aboutToVisible;
	QTimer 	*showHideTimer;
	QStringList focusRoles, focusNoRoles;
	int 	autoShowDelay;
private slots:
	void showHideTimeout();
#endif

#ifdef WITH_SONNET
private:
    Sonnet::Speller *speller;
    //Sonnet::GuessLanguage *guesser;
    
    ControlButton *suggestMenu;
    MyListWidget/*QTableWidget*/ * suggestTable;

    QClipboard *clipboard;
    bool fromClipboard;

    void updateSuggestedMenu();
    void resetSuggestedMenu();
    QIcon newIcon;

private slots:
    //void suggested(int,int);
    void suggested(QListWidgetItem *);
    void checkClipboard(QClipboard::Mode);
    void useSuggestSlot(bool);
    void showSuggestTable(bool);
#endif
};

class KbdTray : public QSystemTrayIcon
{
	Q_OBJECT
public:
	KbdTray(MainWidget* parent = nullptr);
	~KbdTray();

private slots:
	void triggerParentWidgetVisibility(QSystemTrayIcon::ActivationReason reason);

private:
	MainWidget *p;
};

#endif
