/*
 * This file is part of the Kvkbd project.
 * Copyright (C) 2007-2008 Todor Gyumyushev <yodor@developer.bg>
 * Copyright (C) 2008 Guillaume Martres <smarter@ubuntu.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef VBUTTON_H
#define VBUTTON_H

#include <QPushButton>

class QString;
class QRect;
class QTimer;
class QSequentialAnimationGroup;

class VButton : public QPushButton
{
    Q_OBJECT
public:
    explicit VButton(QWidget *parent = nullptr);
    ~VButton();
    void setKeyCode(unsigned int keycode);
    unsigned int getKeyCode();
    void setText(const QString& input, const QString& shiftInput = QString());
    QString getText()      const { return text; }
    QString getShiftText() const { return shiftText; }
    void setShiftText(const QString& key);
    void setKeyCap(bool shiftPressed, bool capsPressed);

    void reposition(int width, int height);
    void res();
    static double pw;
    static double ph;

    void startTimer();
    void stopTimer();

private:
    QTimer *keyTimer;
    QPoint timerStartPos;
    bool rightClicked;
    bool pressed;
    bool grabbed;
    QSequentialAnimationGroup *animGroup;

protected:
    unsigned int keycode;
    QString text;
    QString shiftText;
    QRect originalSize;

public slots:
    void sendKey();

protected:
    void mousePressEvent(QMouseEvent *e) override;
    void mouseReleaseEvent(QMouseEvent *e) override;
	void mouseMoveEvent(QMouseEvent * e) override; //alex
	void leaveEvent(QEvent * e) override; //alex

protected slots:
    void repeatKey();

signals:
    void keyClick(unsigned int keycode);

};

#endif // VBUTTON_H
